from .config_enum import ConfigCategory,ConfigEnum
from .config import StrConfig,BoolConfig,get_hconfigs,hconfig,set_hconfig,add_or_update_config,bulk_register_configs

from .parent_domain import ParentDomain,add_or_update_parent_domains,bulk_register_parent_domains
from .domain import Domain,DomainType,ShowDomain,get_domain,get_current_proxy_domains,get_panel_domains,get_proxy_domains,get_proxy_domains_db,get_hdomains,hdomain,add_or_update_domain,bulk_register_domains
from .proxy import Proxy,ProxyL3,ProxyCDN,ProxyProto,ProxyTransport,add_or_update_proxy,bulk_register_proxies
from .user import User,UserMode,is_user_active,remaining_days,days_to_reset,user_by_id,user_by_uuid,add_or_update_user,user_should_reset,add_or_update_user,bulk_register_users,UserDetail
from .admin import AdminUser,AdminMode,get_super_admin_secret,get_admin_user_db,get_admin_by_uuid,add_or_update_admin,bulk_register_admins
from .child import Child
from .usage import DailyUsage,get_daily_usage_stats




from flask import abort, jsonify, request
from flask_restful import Resource
# from flask_simplelogin import login_required
import datetime
from hiddifypanel.panel.database import db
from hiddifypanel.models import *
from urllib.parse import urlparse
from hiddifypanel.panel import hiddify, usage
from .tgbot import bot


class SendMsgResource(Resource):
    def post(self):

        if not hconfig(ConfigEnum.telegram_bot_token):
            abort(400)

        msg = request.json
        users = User.query.filter(User.telegram_id != None)
        id = msg['id']
        if type(id) == int or id.isnumeric():
            users = [users.filter(User.id == int(msg['id'])).first() or abort(403)]
        elif id == 'all':
            users = users.all()
        else:
            users = users.all()
            if id == 'expired':
                users = [u for u in users if not is_user_active(u)]
            elif id == 'active':
                users = [u for u in users if is_user_active(u)]
            elif id == 'offline 1h':
                h1 = datetime.datetime.now()-datetime.timedelta(hours=1)
                users = [u for u in users if is_user_active(u) and u.last_online < h1]
            elif id == 'offline 1d':
                d1 = datetime.datetime.now()-datetime.timedelta(hours=24)
                users = [u for u in users if is_user_active(u) and u.last_online < d1]

            elif id == 'offline 1w':
                d7 = datetime.datetime.now()-datetime.timedelta(days=7)
                users = [u for u in users if is_user_active(u) and u.last_online < d7]

        res = {}
        for user in users:
            try:
                from hiddifypanel.panel.commercial.telegrambot import Usage
                keyboard = Usage.user_keyboard(user.uuid)
                txt = msg['text']+"\n\n"+Usage.get_usage_msg(user.uuid)
                print('sending to ', user)
                bot.send_message(user.telegram_id, txt, reply_markup=keyboard)
            except Exception as e:
                import traceback
                res[user.uuid] = {'name': user.name, 'error': f'{e}'}
        if len(res) == 0:
            return {'msg': "success"}
        else:
            return {'msg': 'error', 'res': res}



#!/usr/bin/env python3
from urllib.parse import urlparse

from hiddifypanel.panel.database import db
import uuid
from flask_babelex import gettext as _
from flask_bootstrap import SwitchField
# from flask_babelex import gettext as _
import wtforms as wtf
from flask_wtf import FlaskForm
import pathlib
from hiddifypanel.models import *
from hiddifypanel.panel import hiddify
from datetime import datetime, timedelta, date
import os
import sys
import json
import urllib.request
import subprocess
import re
from hiddifypanel.panel import hiddify
from flask import current_app, render_template, request, Response, Markup, url_for, jsonify, redirect, g
from hiddifypanel.panel.hiddify import flash
from flask_wtf.file import FileField, FileRequired
import json

from flask_classful import FlaskView, route


class Backup(FlaskView):

    @hiddify.super_admin
    def index(self):
        return render_template('backup.html', restore_form=get_restore_form())

    # @route("/backupfile")
    @hiddify.super_admin
    def backupfile(self):
        response = jsonify(
            hiddify.dump_db_to_dict()
        )
        o = urlparse(request.base_url)
        domain = o.hostname
        response.headers.add('Content-disposition', f'attachment; filename=hiddify-{domain}-{datetime.now()}.json')

        return response

    @hiddify.super_admin
    def post(self):
        restore_form = get_restore_form()

        if restore_form.validate_on_submit():
            file = restore_form.restore_file.data
            json_data = json.load(file)

            hiddify.set_db_from_json(json_data,
                                     set_users=restore_form.enable_user_restore.data,
                                     set_domains=restore_form.enable_domain_restore.data,
                                     set_settings=restore_form.enable_config_restore.data,
                                     override_unique_id=False,
                                     override_child_id=None
                                     )

            from flask_babel import refresh
            refresh()
            return redirect(f'/{hconfig(ConfigEnum.proxy_path)}/{g.admin.uuid}/admin/actions/reinstall2/', code=302)
            from . import Actions
            action = Actions()
            return action.reinstall(complete_install=True, domain_changed=True)
            # hiddify.flash_config_success(full_install=True)
        else:
            flash(_('Config file is incorrect'))
        return render_template('backup.html', restore_form=restore_form)


def get_restore_form(empty=False):
    class RestoreForm(FlaskForm):
        restore_file = FileField(_("Restore File"), description=_("Restore File Description"), validators=[FileRequired()])
        enable_config_restore = SwitchField(_("Restore Settings"), description=_("Restore Settings description"), default=False)
        enable_user_restore = SwitchField(_("Restore Users"), description=_("Restore Users description"), default=False)
        enable_domain_restore = SwitchField(_("Restore Domain"), description=_("Restore Domain description"), default=False)
        submit = wtf.fields.SubmitField(_('Submit'))

    return RestoreForm(None) if empty else RestoreForm()



from flask_admin.contrib import sqla
from hiddifypanel.panel.database import db
from wtforms.validators import Regexp
from hiddifypanel.models import *
from wtforms.validators import Regexp, ValidationError
from .adminlte import AdminLTEModelView
from flask_babelex import gettext as __
from flask_babelex import lazy_gettext as _
from hiddifypanel.panel import hiddify, cf_api, custom_widgets

from flask import Markup
from flask import Flask, g, flash, url_for
from flask_sqlalchemy import SQLAlchemy
from flask_admin import Admin
from flask_admin.contrib.sqla import ModelView
from wtforms import SelectMultipleField
import re

from wtforms.widgets import ListWidget, CheckboxInput
from sqlalchemy.orm import backref
# Define a custom field type for the related domains
from flask_admin.form.fields import Select2TagsField, Select2Field


# class ConfigDomainsField(SelectField):
#     def __init__(self, label=None, validators=None,*args, **kwargs):
#         kwargs.pop("allow_blank")
#         super().__init__(label, validators,*args, **kwargs)
#         self.choices=[(d.id,d.domain) for d in Doamin.query.filter(Domain.sub_link_only!=True).all()]


class DomainAdmin(AdminLTEModelView):
    column_hide_backrefs = False

    list_template = 'model/domain_list.html'
    # edit_modal = True
    form_overrides = {'mode': custom_widgets.EnumSelectField}
    form_widget_args = {
        'description': {
            'rows': 100,
            'style': 'font-family: monospace; direction:ltr'
        }
    }
    column_descriptions = dict(
        domain=_("domain.description"),
        mode=_("Direct mode means you want to use your server directly (for usual use), CDN means that you use your server on behind of a CDN provider."),
        cdn_ip=_("config.cdn_forced_host.description"),
        show_domains=_('You can select the configs with which domains show be shown in the user area. If you select all, automatically, all the new domains will be added for each users.'),
        alias=_('The name shown in the configs for this domain.'),
        servernames=_('config.reality_server_names.description'),
        sub_link_only=_('This can be used for giving your users a permanent non blockable links.'),
        grpc=_('gRPC is a H2 based protocol. Maybe it is faster for you!')
    )

    # create_modal = True
    can_export = False
    form_widget_args = {'show_domains': {'class': 'form-control ltr'}}
    form_args = {
        'mode': {'enum': DomainType},
        'show_domains': {
            'query_factory': lambda: Domain.query.filter(Domain.sub_link_only == False),
        },
        'domain': {
            'validators': [Regexp(r'^(\*\.)?([A-Za-z0-9\-\.]+\.[a-zA-Z]{2,})$', message=__("Should be a valid domain"))]
        },
        "cdn_ip": {
            'validators': [Regexp(r"(((((25[0-5]|(2[0-4]|1\d|[1-9]|)\d).){3}(25[0-5]|(2[0-4]|1\d|[1-9]|)\d))|^([A-Za-z0-9\-\.]+\.[a-zA-Z]{2,}))[ \t\n,;]*\w{3}[ \t\n,;]*)*", message=__("Invalid IP or domain"))]
        },
        "servernames": {
            'validators': [Regexp(r"^([\w-]+\.)+[\w-]+(,\s*([\w-]+\.)+[\w-]+)*$", re.IGNORECASE, _("Invalid REALITY hostnames"))]
        }
    }
    column_list = ["domain", "alias", "mode", "domain_ip", "show_domains"]
    # column_editable_list=["domain"]
    # column_filters=["domain","mode"]
    # form_excluded_columns=['work_with']
    column_searchable_list = ["domain", "mode"]
    column_labels = {
        "domain": _("domain.domain"),
        'sub_link_only': _('Only for sublink?'),
        "mode": _("domain.mode"),
        "cdn_ip": _("config.cdn_forced_host.label"),
        'domain_ip': _('domain.ip'),
        'servernames': _('config.reality_server_names.label'),
        'show_domains': _('Show Domains'),
        'alias': _('Alias'),
        'grpc': _('gRPC')
    }

    form_columns = ['mode', 'domain', 'alias', 'servernames', 'grpc', 'cdn_ip', 'show_domains']

    def _domain_admin_link(view, context, model, name):
        if model.mode == DomainType.fake:
            return Markup(f"<span class='badge'>{model.domain}</span>")
        d = model.domain
        if "*" in d:
            d = d.replace("*", hiddify.get_random_string(5, 15))
        admin_link = f'https://{d}{hiddify.get_admin_path()}'
        return Markup(f'<div class="btn-group"><a href="{admin_link}" class="btn btn-xs btn-secondary">'+_("admin link")+f'</a><a href="{admin_link}" class="btn btn-xs btn-info ltr" target="_blank">{model.domain}</a></div>')

    def _domain_ip(view, context, model, name):
        dip = hiddify.get_domain_ip(model.domain)
        myip = hiddify.get_ip(4)
        if myip == dip and model.mode == DomainType.direct:
            badge_type = ''
        elif dip and model.mode != DomainType.direct and myip != dip:
            badge_type = 'warning'
        else:
            badge_type = 'danger'
        res = f'<span class="badge badge-{badge_type}">{dip}</span>'
        if model.sub_link_only:
            res += f'<span class="badge badge-success">{_("SubLink")}</span>'
        return Markup(res)

    def _show_domains_formater(view, context, model, name):
        if not len(model.show_domains):
            return _("All")
        else:
            return Markup(" ".join([hiddify.get_domain_btn_link(d) for d in model.show_domains]))

    column_formatters = {
        'domain_ip': _domain_ip,
        'domain': _domain_admin_link,
        'show_domains': _show_domains_formater
    }

    def search_placeholder(self):
        return f"{_('search')} {_('domain.domain')} {_('domain.mode')}"

    # def on_form_prefill(self, form, id):
        # Get the Domain object being edited
        # domain = self.session.query(Domain).get(id)

        # Pre-select the related domains in the checkbox list
        # form.show_domains = [d.id for d in Domain.query.all()]

    def on_model_change(self, form, model, is_created):
        model.domain = model.domain.lower()

        configs = get_hconfigs()
        for c in configs:
            if "domain" in c and c not in [ConfigEnum.decoy_domain, ConfigEnum.reality_fallback_domain] and c.category != 'hidden':
                if model.domain == configs[c]:
                    raise ValidationError(_("You have used this domain in: ")+_(f"config.{c}.label"))

        for td in Domain.query.filter(Domain.mode == DomainType.reality, Domain.domain != model.domain).all():
            print(td)
            if td.servernames and (model.domain in td.servernames.split(",")):
                raise ValidationError(_("You have used this domain in: ")+_(f"config.reality_server_names.label")+" in " + d.domain)
        myip = hiddify.get_ip(4)
        myipv6 = hiddify.get_ip(6)

        if "*" in model.domain and model.mode not in [DomainType.cdn, DomainType.auto_cdn_ip]:
            raise ValidationError(_("Domain can not be resolved! there is a problem in your domain"))

        skip_check = "*" in model.domain
        if hconfig(ConfigEnum.cloudflare) and model.mode != DomainType.fake:
            try:
                proxied = model.mode in [DomainType.cdn, DomainType.auto_cdn_ip]
                cf_api.add_or_update_domain(model.domain, myip, "A", proxied=proxied)
                if myipv6:
                    cf_api.add_or_update_domain(model.domain, myipv6, "AAAA", proxied=proxied)

                skip_check = True
            except Exception as e:
                # raise e
                raise ValidationError(__("Can not connect to Cloudflare.")+f' {e}')
        # elif model.mode==DomainType.auto_cdn_ip:
        if model.alias and not model.alias.replace("_", "").isalnum():
            flash(__("Using alias with special charachters may cause problem in some clients like FairVPN."), 'warning')
        #     raise ValidationError(_("You have to add your cloudflare api key to use this feature: "))

        dip = hiddify.get_domain_ip(model.domain)
        if model.sub_link_only:
            if dip == None:
                raise ValidationError(_("Domain can not be resolved! there is a problem in your domain"))
        elif not skip_check:
            if dip == None:
                raise ValidationError(_("Domain can not be resolved! there is a problem in your domain"))
            domain_ip_is_same_to_panel = myip == dip or dip == myipv6

            if model.mode == DomainType.direct and not domain_ip_is_same_to_panel:
                raise ValidationError(_("Domain IP=%(domain_ip)s is not matched with your ip=%(server_ip)s which is required in direct mode", server_ip=myip, domain_ip=dip))

            if domain_ip_is_same_to_panel and model.mode in [DomainType.cdn, DomainType.relay, DomainType.fake, DomainType.auto_cdn_ip]:
                raise ValidationError(_("In CDN mode, Domain IP=%(domain_ip)s should be different to your ip=%(server_ip)s", server_ip=myip, domain_ip=dip))

            # if model.mode in [DomainType.ss_faketls, DomainType.telegram_faketls]:
            #     if len(Domain.query.filter(Domain.mode==model.mode and Domain.id!=model.id).all())>0:
            #         ValidationError(f"another {model.mode} is exist")
        model.domain = model.domain.lower()
        if model.mode == DomainType.direct and model.cdn_ip:
            raise ValidationError(f"Specifying CDN IP is only valid for CDN mode")

        if model.mode == DomainType.fake and not model.cdn_ip:
            model.cdn_ip = myip

        # if model.mode==DomainType.fake and model.cdn_ip!=myip:
        #     raise ValidationError(f"Specifying CDN IP is only valid for CDN mode")

        # work_with_ids = form.work_with.data
        # print(work_with_ids)
        # # Update the many-to-many relationship
        if len(model.show_domains) == Domain.query.count():
            model.show_domains = []
        # if model.alias and not g.is_commercial:
            #     model.alias= "@hiddify "+model.alias
        # model.work_with = self.session.query(Domain).filter(
        #     Domain.id.in_(work_with_ids)).all()

        if model.mode == DomainType.reality:
            model.servernames = (model.servernames or model.domain).lower()

            for v in [model.domain, model.servernames]:

                for d in v.split(","):
                    if not d:
                        continue

                    if not hiddify.is_domain_reality_friendly(d):
                        # flash(_("Domain is not REALITY friendly!")+" "+d,'error')
                        # return render_template('config.html', form=form)
                        raise ValidationError(_("Domain is not REALITY friendly!")+" "+d)

                    hiddify.debug_flash_if_not_in_the_same_asn(d)

            for d in model.servernames.split(","):
                if not hiddify.fallback_domain_compatible_with_servernames(model.domain, d):
                    raise ValidationError(_("REALITY Fallback domain is not compaitble with server names!")+" "+d+" != "+model.domain)

        if (model.cdn_ip):
            from hiddifypanel.panel import clean_ip
            try:
                clean_ip.get_clean_ip(model.cdn_ip)
            except:
                raise ValidationError(_("Error in auto cdn format"))

        old_db_domain = get_domain(model.domain)
        if is_created or not old_db_domain or old_db_domain.mode != model.mode:
            # return hiddify.reinstall_action(complete_install=False, domain_changed=True)
            hiddify.flash_config_success(restart_mode='apply', domain_changed=True)

    # def after_model_change(self,form, model, is_created):
    #     if model.show_domains.count==0:
    #         db.session.bulk_save_objects(ShowDomain(model.id,model.id))

    def on_model_delete(self, model):
        if len(Domain.query.all()) <= 1:
            raise ValidationError(f"at least one domain should exist")
        hiddify.flash_config_success(restart_mode='apply', domain_changed=True)

    def after_model_delete(self, model):
        # if hconfig(ConfigEnum.parent_panel):
        #     hiddify_api.sync_child_to_parent()
        pass

    def after_model_change(self, form, model, is_created):
        # if hconfig(ConfigEnum.parent_panel):
        #     hiddify_api.sync_child_to_parent()
        pass

    def is_accessible(self):
        return g.admin.mode in [AdminMode.admin, AdminMode.super_admin]

    # def form_choices(self, field, *args, **kwargs):
    #     if field.type == "Enum":
    #         return [(enum_value.name, _(enum_value.name)) for enum_value in field.type.__members__.values()]
    #     return super().form_choices(field, *args, **kwargs)

    # @property
    # def server_ips(self):
    #     return hiddify.get_ip(4)



from sqlalchemy_serializer import SerializerMixin
from flask import Flask
from flask_sqlalchemy import SQLAlchemy
from flask_admin import Admin
from flask_admin.contrib.sqla import ModelView
from wtforms import SelectMultipleField
from sqlalchemy.orm import backref
from flask import Markup
from dateutil import relativedelta
from hiddifypanel.panel.database import db
import enum
import datetime
import uuid as uuid_mod
from enum import auto
from strenum import StrEnum


class ProxyTransport(StrEnum):
    grpc = auto()
    XTLS = auto()
    faketls = auto()
    shadowtls = auto()
    restls1_2 = auto()
    restls1_3 = auto()
    # h1=auto()
    WS = auto()
    tcp = auto()
    ssh = auto()
    custom = auto()


class ProxyCDN(StrEnum):
    CDN = auto()
    direct = auto()
    Fake = auto()


class ProxyProto(StrEnum):
    vless = auto()
    trojan = auto()
    vmess = auto()
    ss = auto()
    v2ray = auto()
    ssr = auto()
    ssh = auto()
    tuic = auto()
    hysteria = auto()


class ProxyL3(StrEnum):
    tls = auto()
    tls_h2 = auto()
    tls_h2_h1 = auto()
    h3_quic = auto()
    reality = auto()
    http = auto()
    kcp = auto()
    ssh = auto()
    custom = auto()


class Proxy(db.Model, SerializerMixin):
    id = db.Column(db.Integer, primary_key=True, autoincrement=True)
    child_id = db.Column(db.Integer, db.ForeignKey('child.id'), default=0)
    name = db.Column(db.String(200), nullable=False, unique=True)
    enable = db.Column(db.Boolean, nullable=False)
    proto = db.Column(db.Enum(ProxyProto), nullable=False)
    l3 = db.Column(db.Enum(ProxyL3), nullable=False)
    transport = db.Column(db.Enum(ProxyTransport), nullable=False)
    cdn = db.Column(db.Enum(ProxyCDN), nullable=False)

    def to_dict(d):
        return {
            'name': d.name,
            'enable': d.enable,
            'proto': d.proto,
            'l3': d.l3,
            'transport': d.transport,
            'cdn': d.cdn,
            'child_unique_id': d.child.unique_id if d.child else ''
        }


def add_or_update_proxy(commit=True, child_id=0, **proxy):
    from hiddifypanel.panel import hiddify
    dbproxy = Proxy.query.filter(Proxy.name == proxy['name']).first()
    if not dbproxy:
        dbproxy = Proxy()
        db.session.add(dbproxy)
    dbproxy.enable = proxy['enable']
    dbproxy.name = proxy['name']
    dbproxy.proto = proxy['proto']
    dbproxy.transport = proxy['transport']
    dbproxy.cdn = proxy['cdn']
    dbproxy.l3 = proxy['l3']
    dbproxy.child_id = child_id
    if commit:
        db.session.commit()


def bulk_register_proxies(proxies, commit=True, override_child_id=None):
    from hiddifypanel.panel import hiddify
    for proxy in proxies:
        child_id = override_child_id if override_child_id is not None else hiddify.get_child(proxy.get('child_unique_id', None))
        add_or_update_proxy(commit=False, child_id=child_id, **proxy)



#!/usr/bin/env python3
import pprint
from flask_babelex import gettext as _
import pathlib
from hiddifypanel.models import *

from datetime import datetime, timedelta, date
import os
import sys
import json
import urllib.request
import subprocess
import re
from hiddifypanel.panel import hiddify, usage
from flask import current_app, render_template, request, Response, Markup, url_for
from hiddifypanel.panel.hiddify import flash

from flask_classful import FlaskView, route


class CommercialInfo(FlaskView):

    def index(self):
        return render_template('commercial_info.html')



import datetime
import uuid as uuid_mod

from sqlalchemy_serializer import SerializerMixin
from dateutil import relativedelta
from hiddifypanel.panel.database import db
from enum import auto
from strenum import StrEnum
from flask import g, abort


class AdminMode(StrEnum):
    """
    The "UserMode" class is an enumeration that defines five possible modes: "no_reset", "monthly", "weekly",
    "daily", and "disable". These modes represent different settings that can be applied to a user account,
    such as the frequency at which data is reset or whether the account is currently disabled. The class is
    implemented using the "StrEnum" base class and the "auto()" function to generate unique values for each mode.
    """
    super_admin = auto()
    admin = auto()
    agent = auto()


class AdminUser(db.Model, SerializerMixin):
    """
    This is a model class for a user in a database that includes columns for their ID, UUID, name, online status,
    account expiration date, usage limit, package days, mode, start date, current usage, last reset time, and comment.
    """
    id = db.Column(db.Integer, primary_key=True, autoincrement=True)
    uuid = db.Column(db.String(36), default=lambda: str(uuid_mod.uuid4()), nullable=False, unique=True)
    name = db.Column(db.String(512), nullable=False)
    mode = db.Column(db.Enum(AdminMode), default=AdminMode.agent, nullable=False)
    can_add_admin = db.Column(db.Boolean, default=False, nullable=False)
    max_users = db.Column(db.Integer, default=100, nullable=False)
    max_active_users = db.Column(db.Integer, default=100, nullable=False)
    comment = db.Column(db.String(512))
    telegram_id = db.Column(db.String(512))
    users = db.relationship('User', backref='admin')
    usages = db.relationship('DailyUsage', backref='admin')
    parent_admin_id = db.Column(db.Integer, db.ForeignKey('admin_user.id'), default=1)
    parent_admin = db.relationship('AdminUser', remote_side=[id], backref='sub_admins')

    def recursive_users_query(self):
        from .user import User
        admin_ids = self.recursive_sub_admins_ids()
        return User.query.filter(User.added_by.in_(admin_ids))

    def can_have_more_users(admin):
        if admin.mode == AdminMode.super_admin:
            return True
        users_count = admin.recursive_users_query().count()
        if admin.max_users < users_count:
            return False
        if users_count <= admin.max_active_users:
            return True

        actives = [u for u in admin.recursive_users_query().all() if u.is_active]
        return len(actives) <= admin.max_active_users

    def recursive_sub_admins_ids(self, depth=20, seen=None):
        if seen is None:
            seen = set()
        sub_admin_ids = []
        if self.id not in seen:
            sub_admin_ids.append(self.id)
            seen.add(self.id)
        if depth > 0:
            for sub_admin in self.sub_admins:
                sub_admin_ids += sub_admin.recursive_sub_admins_ids(depth-1, seen=seen)
        return sub_admin_ids

    def __str__(self):
        return str(self.name)

    def to_dict(admin):
        return {
            'name': admin.name,
            'comment': admin.comment,
            'uuid': admin.uuid,
            'mode': admin.mode,
            'can_add_admin': admin.can_add_admin,
            'parent_admin_uuid': admin.parent_admin.uuid if admin.parent_admin else None,
            'telegram_id': admin.telegram_id
        }


def get_super_admin_secret():
    admin = AdminUser.query.filter(AdminUser.mode == AdminMode.super_admin).first()
    if not admin:
        db.session.add(AdminUser(mode=AdminMode.super_admin, name="Owner"))
        db.session.commit()
        admin = AdminUser.query.filter(AdminUser.mode == AdminMode.super_admin).first()

    admin_secret = admin.uuid
    return admin_secret


def get_admin_user_db(uuid):
    return AdminUser.query.filter(AdminUser.uuid == uuid).first()


def add_or_update_admin(commit=True, **admin):
    # if not is_valid():return
    dbuser = AdminUser.query.filter(AdminUser.uuid == admin['uuid']).first()

    if not dbuser:
        dbuser = AdminUser(uuid=admin['uuid'])
        # if not is_valid():
        #     return
        db.session.add(dbuser)
    dbuser.name = admin['name']
    if dbuser.id != 1:
        parent = admin.get('parent_admin_uuid')
        if parent == admin['uuid'] or not parent:
            parent_admin = current_admin_or_owner()
        else:
            parent_admin = get_admin_by_uuid(parent, create=True)
        dbuser.parent_admin_id = parent_admin.id

    dbuser.comment = admin.get('comment', '')
    dbuser.mode = admin.get('mode', AdminMode.agent)
    dbuser.telegram_id = admin.get('telegram_id')
    dbuser.can_add_admin = admin.get('can_add_admin') == True

    # dbuser.last_online=user.get('last_online','')
    if commit:
        db.session.commit()


def get_admin_by_uuid(uuid, create=False):
    admin = AdminUser.query.filter(AdminUser.uuid == uuid).first()
    # print(admin)
    if create and not admin:

        # print(parent_id)
        # raise (parent_id)
        dbuser = AdminUser(uuid=uuid, name="unknown", parent_admin_id=current_admin_or_owner().id)
        db.session.add(dbuser)
        db.session.commit()
        admin = AdminUser.query.filter(AdminUser.uuid == uuid).first()
    return admin


def bulk_register_admins(users=[], commit=True):
    for u in users:
        add_or_update_admin(commit=False, **u)
    if commit:
        db.session.commit()


def current_admin_or_owner():
    if g and g.admin:
        return g.admin
    return AdminUser.query.filter(AdminUser.id == 1).first()



import flask_babel
import flask_babelex
from hiddifypanel.models import BoolConfig, StrConfig, ConfigEnum, hconfig, ConfigCategory
from wtforms.validators import Regexp, ValidationError
from flask_babelex import lazy_gettext as _
# from flask_babelex import gettext as _
import wtforms as wtf
from flask_wtf import FlaskForm, CSRFProtect
from flask_bootstrap import SwitchField

from flask_admin.base import expose
# from gettext import gettext as _
from hiddifypanel.panel.hiddify import get_random_domains

import re
from flask import render_template, current_app, Markup, url_for, abort, g
from hiddifypanel.panel.hiddify import flash
from hiddifypanel.models import *
from hiddifypanel.panel.database import db
from wtforms.fields import *
from flask_classful import FlaskView
from hiddifypanel.panel import hiddify, custom_widgets


class SettingAdmin(FlaskView):

    @hiddify.super_admin
    def index(self):
        form = get_config_form()
        return render_template('config.html', form=form)

    @hiddify.super_admin
    def post(self):
        set_hconfig(ConfigEnum.first_setup, False)
        form = get_config_form()
        reset_action = None
        if form.validate_on_submit():

            boolconfigs = BoolConfig.query.filter(BoolConfig.child_id == 0).all()
            bool_types = {c.key: 'bool' for c in boolconfigs}
            old_configs = get_hconfigs()
            changed_configs = {}
            for cat, vs in form.data.items():  # [c for c in ConfigEnum]:

                if type(vs) is dict:
                    for k in ConfigEnum:
                        if k not in vs:
                            continue
                        v = vs[k]
                        if k.type() == str:
                            if "_domain" in k or "_fakedomain" in k:
                                v = v.lower()

                            if "port" in k:
                                for p in v.split(","):
                                    for k2, v2 in vs.items():
                                        if "port" in k2 and k != k2 and p in v2:
                                            flash(_("Port is already used! in")+f" {k2} {k}", 'error')
                                            return render_template('config.html', form=form)
                            if k == ConfigEnum.parent_panel and v != '':
                                # v=(v+"/").replace("/admin",'')
                                v = re.sub("(/admin/.*)", "/", v)

                                # try:
                                #     # if hiddify_api.sync_child_to_parent(v)['status'] != 200:
                                #     #     flash(_("Can not connect to parent panel!"), 'error')
                                #     #     return render_template('config.html', form=form)
                                # except:
                                #     flash(_("Can not connect to parent panel!"), 'error')
                                #     return render_template('config.html', form=form)
                            StrConfig.query.filter(StrConfig.key == k, StrConfig.child_id == 0).first().value = v
                        if old_configs[k] != v:
                            changed_configs[k] = v

                # print(cat,vs)

            merged_configs = {**old_configs, **changed_configs}
            # for k in [ConfigEnum.reality_server_names,ConfigEnum.reality_fallback_domain]:
            #     v=merged_configs[k]
            #     for d in v.split(","):
            #         if not d:continue
            #         if not hiddify.is_domain_reality_friendly(d):
            #             flash(_("Domain is not REALITY friendly!")+" "+d,'error')
            #             return render_template('config.html', form=form)
            #         hiddify.debug_flash_if_not_in_the_same_asn(d)
            # fallback=merged_configs[ConfigEnum.reality_fallback_domain]
            # for d in merged_configs[ConfigEnum.reality_server_names].split(","):
            #     if not hiddify.fallback_domain_compatible_with_servernames(fallback, d):
            #         flash(_("REALITY Fallback domain is not compaitble with server names!")+" "+d+" != "+fallback,'error')
            #         return render_template('config.html', form=form)
            for k, v in changed_configs.items():
                set_hconfig(k, v, 0, False)
            db.session.commit()
            flask_babel.refresh()
            flask_babelex.refresh()

            from hiddifypanel.panel.commercial.telegrambot import register_bot
            register_bot()
            # if hconfig(ConfigEnum.parent_panel):
            #     hiddify_api.sync_child_to_parent()
            reset_action = hiddify.check_need_reset(old_configs)
            # if :
            #     return reset_action

            if old_configs[ConfigEnum.admin_lang] != hconfig(ConfigEnum.admin_lang):
                form = get_config_form()
        else:
            flash(_('config.validation-error'), 'danger')

        return reset_action or render_template('config.html', form=form)

        # # form=HelloForm()
        # # # return render('config.html',form=form)
        # # return render_template('config.html',form=HelloForm())
        # form=get_config_form()
        # return render_template('config.html',form=form)

    def get_babel_string(self):
        res = ""
        strconfigs = StrConfig.query.all()
        boolconfigs = BoolConfig.query.all()
        bool_types = {c.key: 'bool' for c in boolconfigs}

        configs = [*boolconfigs, *strconfigs]
        for cat in ConfigCategory:
            if cat == 'hidden':
                continue

            cat_configs = [c for c in configs if c.key.category() == cat]

            for c in cat_configs:
                res += f'{{{{_("config.{c.key}.label")}}}} {{{{_("config.{c.key}.description")}}}}'

            res += f'{{{{_("config.{cat}.label")}}}}{{{{_("config.{cat}.description")}}}}'
        for c in ConfigEnum:
            res += f'{{{{_("config.{c}.label")}}}} {{{{_("config.{c}.description")}}}}'
        return res


def get_config_form():

    strconfigs = StrConfig.query.filter(StrConfig.child_id == 0).all()
    boolconfigs = BoolConfig.query.filter(BoolConfig.child_id == 0).all()
    bool_types = {c.key: 'bool' for c in boolconfigs}

    configs = [*boolconfigs, *strconfigs]
    configs_key = {k.key: k for k in configs}
    # categories=sorted([ c for c in {c.key.category():1 for c in configs}])
    # dict_configs={cat:[c for c in configs if c.category==cat] for cat in categories}

    class DynamicForm(FlaskForm):
        pass
    is_parent = hconfig(ConfigEnum.is_parent)

    for cat in ConfigCategory:
        if cat == 'hidden':
            continue

        cat_configs = [c for c in ConfigEnum if c.category() == cat and (not is_parent or c.show_in_parent())]

        if len(cat_configs) == 0:
            continue

        class CategoryForm(FlaskForm):
            description_for_fieldset = wtf.fields.TextAreaField("", description=_(f'config.{cat}.description'), render_kw={"class": "d-none"})
        for c2 in cat_configs:
            if not (c2 in configs_key):
                continue
            c = configs_key[c2]
            extra_info = ''
            if c.key in bool_types:
                field = SwitchField(_(f'config.{c.key}.label'), default=c.value, description=_(f'config.{c.key}.description'))
            elif c.key == ConfigEnum.core_type:
                field = wtf.fields.SelectField(_(f"config.{c.key}.label"), choices=[("xray", _("Xray")), ("singbox", _(
                    "SingBox"))], description=_(f"config.{c.key}.description"), default=hconfig(c.key))
            elif c.key == ConfigEnum.warp_mode:
                field = wtf.fields.SelectField(_(f"config.{c.key}.label"), choices=[("disable", _("Disable")), ("all", _("All")), ("custom", _(
                    "Only Blocked and Local websites"))], description=_(f"config.{c.key}.description"), default=hconfig(c.key))

            elif c.key == ConfigEnum.lang or c.key == ConfigEnum.admin_lang:
                field = wtf.fields.SelectField(_(f"config.{c.key}.label"), choices=[("en", _("lang.en")), ("fa", Markup(_("lang.fa"))), ("zh", _(
                    "lang.zh")), ("pt", _("lang.pt"))], description=_(f"config.{c.key}.description"), default=hconfig(c.key))
            elif c.key == ConfigEnum.country:
                field = wtf.fields.SelectField(_(f"config.{c.key}.label"), choices=[("ir", _("Iran")), ("zh", _(
                    "China")), ("other", _("Others"))], description=_(f"config.{c.key}.description"), default=hconfig(c.key))
            elif c.key == ConfigEnum.package_mode:
                field = wtf.fields.SelectField(_(f"config.{c.key}.label"), choices=[("release", _("Release")),
                                               ("develop", _("Develop"))], description=_(f"config.{c.key}.description"), default=hconfig(c.key))
            elif c.key == ConfigEnum.utls:
                field = wtf.fields.SelectField(_(f"config.{c.key}.label"), choices=[("none", "None"), ("chrome", "Chrome"), ("edge", "Edge"), ("ios", "iOS"), ("android", "Android"), (
                    "safari", "Safari"), ("firefox", "Firefox"), ('random', 'random'), ('randomized', 'randomized')], description=_(f"config.{c.key}.description"), default=hconfig(c.key))
            elif c.key == ConfigEnum.telegram_lib:
                # if hconfig(ConfigEnum.telegram_lib)=='python':
                #     continue6
                libs = [("python", _("lib.telegram.python")), ("tgo", _("lib.telegram.go")), ("orig", _("lib.telegram.orignal")), ("erlang", _("lib.telegram.erlang"))]
                field = wtf.fields.SelectField(_("config.telegram_lib.label"), choices=libs, description=_(
                    "config.telegram_lib.description"), default=hconfig(ConfigEnum.telegram_lib))
            elif c.key == ConfigEnum.branding_freetext:
                validators = [wtf.validators.Length(max=2048)]
                render_kw = {'class': "ltr", 'maxlength': 2048}
                field = custom_widgets.CKTextAreaField(_(f'config.{c.key}.label'), validators, default=c.value, description=_(f'config.{c.key}.description'), render_kw=render_kw)
            else:
                render_kw = {'class': "ltr"}
                validators = []
                if c.key == ConfigEnum.domain_fronting_domain:
                    validators.append(wtf.validators.Regexp("^([A-Za-z0-9\-\.]+\.[a-zA-Z]{2,})|$", re.IGNORECASE, _("config.Invalid domain")))
                    validators.append(hiddify.validate_domain_exist)
                elif '_domain' in c.key or "_fakedomain" in c.key:
                    validators.append(wtf.validators.Regexp("^([A-Za-z0-9\-\.]+\.[a-zA-Z]{2,})$", re.IGNORECASE, _("config.Invalid domain")))
                    validators.append(hiddify.validate_domain_exist)

                    if c.key != ConfigEnum.decoy_domain:
                        validators.append(wtf.validators.NoneOf([d.domain.lower() for d in Domain.query.all()], _("config.Domain already used")))
                        validators.append(wtf.validators.NoneOf([cc.value.lower() for cc in StrConfig.query.filter(StrConfig.child_id == 0).all(
                        ) if cc.key != c.key and "fakedomain" in cc.key and cc.key != ConfigEnum.decoy_domain], _("config.Domain already used")))
                    render_kw['required'] = ""
                    if len(c.value) < 3:
                        c.value = get_random_domains(1)[0]
                # if c.key ==ConfigEnum.reality_short_ids:
                #     extra_info=f" <a target='_blank' href='{url_for('admin.Actions:get_some_random_reality_friendly_domain',test_domain=c.value)}'>"+_('Example Domains')+"</a>"
                # if c.key ==ConfigEnum.reality_server_names:
                #     validators.append(wtf.validators.Regexp("^([\w-]+\.)+[\w-]+(,\s*([\w-]+\.)+[\w-]+)*$",re.IGNORECASE,_("Invalid REALITY hostnames")))
                    # gauge width gate lamp weasel jaguar minute enough few attitude endorse situate usdt trc20 doge bep20 trx doge ltc bnb eth btc bnb
                    # enjoy control list debris chronic few door broken way negative daring life season recipe profit switch bitter casual frame aunt plate brush aerobic display
                if c.key == ConfigEnum.parent_panel:
                    validators.append(wtf.validators.Regexp("()|(http(s|)://([A-Za-z0-9\-\.]+\.[a-zA-Z]{2,})/.*)", re.IGNORECASE, _("Invalid admin link")))
                if c.key == ConfigEnum.telegram_bot_token:
                    validators.append(wtf.validators.Regexp("()|^([0-9]{8,12}:[a-zA-Z0-9_-]{30,40})|$", re.IGNORECASE, _("config.Invalid telegram bot token")))
                if c.key == ConfigEnum.branding_site:
                    validators.append(wtf.validators.Regexp("()|(http(s|)://([A-Za-z0-9\-\.]+\.[a-zA-Z]{2,})/?.*)", re.IGNORECASE, _("config.Invalid brand link")))
                    # render_kw['required']=""

                if 'secret' in c.key:
                    validators.append(wtf.validators.Regexp(
                        "^[0-9a-fA-F]{8}-[0-9a-fA-F]{4}-[0-9a-fA-F]{4}-[0-9a-fA-F]{4}-[0-9a-fA-F]{12}$", re.IGNORECASE, _('config.invalid uuid')))
                    render_kw['required'] = ""

                if c.key == ConfigEnum.proxy_path:
                    validators.append(wtf.validators.Regexp("^[a-zA-Z0-9]*$", re.IGNORECASE, _("config.Invalid proxy path")))
                    render_kw['required'] = ""

                if 'port' in c.key:
                    if c.key in [ConfigEnum.http_ports, ConfigEnum.tls_ports]:
                        validators.append(wtf.validators.Regexp("^(\d+)(,\d+)*$", re.IGNORECASE, _("config.Invalid port")))
                        render_kw['required'] = ""
                    else:
                        validators.append(wtf.validators.Regexp("^(\d+)(,\d+)*$|^$", re.IGNORECASE, _("config.Invalid port")))

                    # validators.append(wtf.validators.Regexp("^(\d+)(,\d+)*$",re.IGNORECASE,_("config.port is required")))

                for val in validators:
                    if hasattr(val, "regex"):
                        render_kw['pattern'] = val.regex.pattern
                        render_kw['title'] = val.message
                if c.key == ConfigEnum.reality_public_key and g.admin.mode in [AdminMode.super_admin]:
                    extra_info = f" <a href='{url_for('admin.Actions:change_reality_keys')}'>{_('Change')}</>"
                field = wtf.fields.StringField(_(f'config.{c.key}.label'), validators, default=c.value,
                                               description=_(f'config.{c.key}.description')+extra_info, render_kw=render_kw)
            setattr(CategoryForm, c.key, field)

        multifield = wtf.fields.FormField(CategoryForm, Markup('<i class="fa-solid fa-plus"></i>&nbsp'+_(f'config.{cat}.label')))

        setattr(DynamicForm, cat,  multifield)

    setattr(DynamicForm, "submit", wtf.fields.SubmitField(_('Submit')))

    return DynamicForm()



from sqlalchemy_serializer import SerializerMixin
from flask import Flask, flash, request
from flask_sqlalchemy import SQLAlchemy
from flask_admin import Admin
from flask_admin.contrib.sqla import ModelView
from wtforms import SelectMultipleField
from urllib.parse import urlparse
from .config_enum import ConfigEnum, ConfigCategory
from .config import hconfig

from sqlalchemy.orm import backref

from flask import Markup
from dateutil import relativedelta
from hiddifypanel.panel.database import db

import enum
import datetime
import uuid as uuid_mod
from enum import auto
from strenum import StrEnum


class DomainType(StrEnum):
    direct = auto()
    sub_link_only = auto()
    cdn = auto()
    auto_cdn_ip = auto()
    relay = auto()
    reality = auto()
    old_xtls_direct = auto()
    worker = auto()
    fake = auto()

    # fake_cdn = "fake_cdn"
    # telegram_faketls = "telegram_faketls"
    # ss_faketls = "ss_faketls"


ShowDomain = db.Table('show_domain',
                      db.Column('domain_id', db.Integer, db.ForeignKey('domain.id'), primary_key=True),
                      db.Column('related_id', db.Integer, db.ForeignKey('domain.id'), primary_key=True)
                      )


class Domain(db.Model, SerializerMixin):
    id = db.Column(db.Integer, primary_key=True, autoincrement=True)
    child_id = db.Column(db.Integer, db.ForeignKey('child.id'), default=0)
    domain = db.Column(db.String(200), nullable=False, unique=False)
    alias = db.Column(db.String(200))
    sub_link_only = db.Column(db.Boolean, nullable=False, default=False)
    mode = db.Column(db.Enum(DomainType), nullable=False, default=DomainType.direct)
    cdn_ip = db.Column(db.Text(2000), nullable=True, default='')
    grpc = db.Column(db.Boolean, nullable=True, default=False)
    servernames = db.Column(db.String(1000), nullable=True, default='')
    # show_all=db.Column(db.Boolean, nullable=True)
    show_domains = db.relationship('Domain', secondary=ShowDomain,
                                   primaryjoin=id == ShowDomain.c.domain_id,
                                   secondaryjoin=id == ShowDomain.c.related_id,
                                   # backref=backref('show_domains', lazy='dynamic')
                                   )

    def __repr__(self):
        return f'{self.domain}'

    def to_dict(d):
        return {
            'domain': d.domain,
            'mode': d.mode,
            'alias': d.alias,
            # 'sub_link_only':d.sub_link_only,
            'child_unique_id': d.child.unique_id if d.child else '',
            'cdn_ip': d.cdn_ip,
            'servernames': d.servernames,
            'grpc': d.grpc,
            'show_domains': [dd.domain for dd in d.show_domains]
        }


def hdomains(mode):
    domains = Domain.query.filter(Domain.mode == mode).all()
    if domains:
        return [d.domain for d in domains]
    return []


def hdomain(mode):
    domains = hdomains(mode)
    if domains:
        return domains[0]
    return None


def get_hdomains():
    return {mode: hdomains(mode) for mode in DomainType}


def get_domain(domain):
    return Domain.query.filter(Domain.domain == domain).first()


def get_panel_domains(always_add_ip=False, always_add_all_domains=False):
    domains = []
    if hconfig(ConfigEnum.is_parent):
        from .parent_domain import ParentDomain
        domains = ParentDomain.query.all()
    else:
        domains = Domain.query.filter(Domain.mode == DomainType.sub_link_only).all()
        if not len(domains) or always_add_all_domains:
            domains = Domain.query.filter(Domain.mode.notin_([DomainType.fake, DomainType.reality])).all()

    if len(domains) == 0 and request:
        domains = [Domain(domain=request.host)]
    if len(domains) == 0 or always_add_ip:
        from hiddifypanel.panel import hiddify
        domains += [Domain(domain=hiddify.get_ip(4))]
    return domains


def get_proxy_domains(domain):
    if hconfig(ConfigEnum.is_parent):
        from hiddifypanel.commercial.parent_domain import ParentDomain
        db_domain = ParentDomain.query.filter(ParentDomain.domain == domain).first() or ParentDomain(domain=domain, show_domains=[])
    else:
        db_domain = Domain.query.filter(Domain.domain == domain).first() or Domain(domain=domain, mode=DomainType.direct, cdn_ip='', show_domains=[])
    return get_proxy_domains_db(db_domain)


def get_proxy_domains_db(db_domain):
    if not db_domain:
        domain = urlparse(request.base_url).hostname
        db_domain = Domain(domain=domain, mode=DomainType.direct, show_domains=[])
        print("no domain")
        flash(_("This domain does not exist in the panel!" + domain))

    return db_domain.show_domains or Domain.query.all()


def get_current_proxy_domains(force_domain=None):
    domain = force_domain or urlparse(request.base_url).hostname
    return get_proxy_domains(domain)


def add_or_update_domain(commit=True, child_id=0, **domain):
    from hiddifypanel.panel import hiddify
    dbdomain = Domain.query.filter(Domain.domain == domain['domain']).first()
    if not dbdomain:
        dbdomain = Domain(domain=domain['domain'])
        db.session.add(dbdomain)
    dbdomain.child_id = child_id

    dbdomain.mode = domain['mode']
    if (str(domain.get('sub_link_only', False)).lower() == 'true'):
        dbdomain.mode = DomainType.sub_link_only
    dbdomain.cdn_ip = domain.get('cdn_ip', '')
    dbdomain.alias = domain.get('alias', '')
    dbdomain.grpc = domain.get('grpc', False)
    dbdomain.servernames = domain.get('servernames', '')
    show_domains = domain.get('show_domains', [])
    dbdomain.show_domains = Domain.query.filter(Domain.domain.in_(show_domains)).all()
    if commit:
        db.session.commit()


def bulk_register_domains(domains, commit=True, remove=False, override_child_id=None):
    from hiddifypanel.panel import hiddify
    child_ids = {}
    for domain in domains:
        child_id = override_child_id if override_child_id is not None else hiddify.get_child(domain.get('child_unique_id', None))
        child_ids[child_id] = 1
        add_or_update_domain(commit=False, child_id=child_id, **domain)
    if remove and len(child_ids):
        dd = {d['domain']: 1 for d in domains}
        for d in Domain.query.filter(Domain.child_id.in_(child_ids)):
            if d.domain not in dd:
                db.session.delete(d)
    if commit:
        db.session.commit()



from flask import g, request, render_template
import enum
from hiddifypanel.models import *
import yaml
import json
from hiddifypanel.panel import hiddify
import random
import re
import datetime
from flask_babelex import gettext as _


def all_proxies(child_id=0):
    all_proxies = hiddify.get_available_proxies(child_id)
    all_proxies = [p for p in all_proxies if p.enable]
    # all_cfg=Proxy.query.filter(Proxy.enable==True).all()
    # if not hconfig(ConfigEnum.domain_fronting_domain):
    #     all_cfg=[c for c in all_cfg if 'Fake' not in c.cdn]
    # if not g.is_cdn:
    #     all_cfg=[c for c in all_cfg if 'CDN' not in c.cdn]
    # if not hconfig(ConfigEnum.ssfaketls_enable):
    #     all_cfg=[c for c in all_cfg if 'faketls' not in c.transport and 'v2ray' not in c.proto]
    # if not hconfig(ConfigEnum.vmess_enable):
    #     all_cfg=[c for c in all_cfg if 'vmess' not in c.proto]

    return all_proxies


def proxy_info(name, mode="tls"):
    return "error"


def pbase64(full_str):
    return full_str
    str = full_str.split("vmess://")[1]
    import base64
    resp = base64.b64encode(f'{str}'.encode("utf-8"))
    return "vmess://"+resp


def make_proxy(proxy: Proxy, domain_db: Domain, phttp=80, ptls=443, pport=None):
    def is_tls():
        return 'tls' in l3 or "reality" in l3
    if type(phttp) == str:
        phttp = int(phttp) if phttp != "None" else None
    if type(ptls) == str:
        ptls = int(ptls) if ptls != "None" else None
    l3 = proxy.l3

    domain = domain_db.domain

    child_id = domain_db.child_id
    hconfigs = get_hconfigs(child_id)
    port = 0

    if is_tls():
        port = ptls
    elif l3 == "http":
        port = phttp
    elif l3 == "kcp":
        port = hconfigs[ConfigEnum.kcp_ports].split(",")[0]
    elif l3 == "tuic":
        port = hconfigs[ConfigEnum.tuic_port].split(",")[0]
    elif l3 == 'ssh':
        port = hconfigs[ConfigEnum.ssh_server_port]
    if pport:
        port = pport

    name = proxy.name
    if not port:
        return {'name': name, 'msg': "port not defined", 'type': 'error', 'proto': proxy.proto}
    if "reality" not in l3 and domain_db.mode == DomainType.reality:
        return {'name': name, 'msg': "reality proxy not in reality domain", 'type': 'debug', 'proto': proxy.proto}

    if "reality" in l3 and domain_db.mode != DomainType.reality:
        return {'name': name, 'msg': "reality proxy not in reality domain", 'type': 'debug', 'proto': proxy.proto}

    if "reality" in l3 and domain_db.grpc and ProxyTransport.grpc != proxy.transport:
        return {'name': name, 'msg': "reality proxy not in reality domain", 'type': 'debug', 'proto': proxy.proto}

    if "reality" in l3 and (not domain_db.grpc) and ProxyTransport.grpc == proxy.transport:
        return {'name': name, 'msg': "reality proxy not in reality domain", 'type': 'debug', 'proto': proxy.proto}

    is_cdn = ProxyCDN.CDN in proxy.cdn
    if is_cdn and domain_db.mode not in [DomainType.cdn, DomainType.auto_cdn_ip, DomainType.worker]:
        # print("cdn proxy not in cdn domain", domain, name)
        return {'name': name, 'msg': "cdn proxy not in cdn domain", 'type': 'debug', 'proto': proxy.proto}
    if not is_cdn and domain_db.mode in [DomainType.cdn, DomainType.auto_cdn_ip, DomainType.worker]:
        # print("not cdn proxy  in cdn domain", domain, name, proxy.cdn)
        return {'name': name, 'msg': "not cdn proxy  in cdn domain", 'type': 'debug', 'proto': proxy.proto}
    if domain_db.mode == DomainType.worker and proxy.transport == ProxyTransport.grpc:
        return {'name': name, 'msg': "worker does not support grpc", 'type': 'debug', 'proto': proxy.proto}
    cdn_forced_host = domain_db.cdn_ip or (domain_db.domain if domain_db.mode != DomainType.reality else hiddify.get_direct_host_or_ip(4))

    alpn = "h2" if proxy.l3 in ['tls_h2', 'reality'] else 'h2,http/1.1' if proxy.l3 == 'tls_h2_h1' else "http/1.1"
    base = {
        'name': name,
        'cdn': is_cdn,
        'mode': "CDN" if is_cdn else "direct",
        'l3': l3,
        'host': domain,
        'port': port,
        'server': cdn_forced_host,
        'sni': domain_db.servernames if is_cdn and domain_db.servernames else domain,
        'uuid': str(g.user_uuid),
        'proto': proxy.proto,
        'transport': proxy.transport,
        'proxy_path': hconfigs[ConfigEnum.proxy_path],
        'alpn': alpn,
        'extra_info': f'{domain_db.alias or domain}',
        'fingerprint': hconfigs[ConfigEnum.utls],
        'allow_insecure': domain_db.mode == DomainType.fake or "Fake" in proxy.cdn,
        'dbe': proxy,
        'dbdomain': domain_db
    }

    if base["proto"] == "trojan" and not is_tls():
        return {'name': name, 'msg': "trojan but not tls", 'type': 'warning', 'proto': proxy.proto}

    if l3 == "http" and ProxyTransport.XTLS in proxy.transport:
        return {'name': name, 'msg': "http and xtls???", 'type': 'warning', 'proto': proxy.proto}
    if l3 == "http" and base["proto"] in ["ss", "ssr"]:
        return {'name': name, 'msg': "http and ss or ssr???", 'type': 'warning', 'proto': proxy.proto}

    if proxy.proto in ProxyProto.vmess:
        base['cipher'] = "chacha20-poly1305"

    if l3 in ['reality']:
        base['reality_short_id'] = random.sample(hconfigs[ConfigEnum.reality_short_ids].split(','), 1)[0]
        # base['flow']="xtls-rprx-vision"
        base['reality_pbk'] = hconfigs[ConfigEnum.reality_public_key]
        if (domain_db.servernames):
            all_servernames = re.split('[ \t\r\n;,]+', domain_db.servernames)
            base['sni'] = random.sample(all_servernames, 1)[0]
            if hconfigs[ConfigEnum.core_type] == "singbox":
                base['sni'] = all_servernames[0]
        else:
            base['sni'] = domain_db.domain

        del base['host']
        if base.get('fingerprint', 'none') != 'none':
            base['fingerprint'] = hconfigs[ConfigEnum.utls]
        # if not domain_db.cdn_ip:
        #     base['server']=hiddify.get_domain_ip(base['server'])

    if "Fake" in proxy.cdn:
        if not hconfigs[ConfigEnum.domain_fronting_domain]:
            return {'name': name, 'msg': "no domain_fronting_domain", 'type': 'debug', 'proto': proxy.proto}
        if l3 == "http" and not hconfigs[ConfigEnum.domain_fronting_http_enable]:
            return {'name': name, 'msg': "no http in domain_fronting_domain", 'type': 'debug', 'proto': proxy.proto}
        if l3 == "tls" and not hconfigs[ConfigEnum.domain_fronting_tls_enable]:
            return {'name': name, 'msg': "no tls in domain_fronting_domain", 'type': 'debug', 'proto': proxy.proto}
        base['server'] = hconfigs[ConfigEnum.domain_fronting_domain]
        base['sni'] = hconfigs[ConfigEnum.domain_fronting_domain]
        # base["host"]=domain
        base['mode'] = 'Fake'

    elif l3 == "http" and not hconfigs[ConfigEnum.http_proxy_enable]:
        return {'name': name, 'msg': "http but http is disabled ", 'type': 'debug', 'proto': proxy.proto}
    path = {
        'vless': f'{hconfigs[ConfigEnum.path_vless]}',
        'trojan': f'{hconfigs[ConfigEnum.path_trojan]}',
        'vmess': f'{hconfigs[ConfigEnum.path_vmess]}',
        'ss': f'{hconfigs[ConfigEnum.path_ss]}',
        'v2ray': f'{hconfigs[ConfigEnum.path_ss]}'
    }

    if base["proto"] in ['v2ray', 'ss', 'ssr']:
        base['cipher'] = 'chacha20-ietf-poly1305'
        if "shadowtls" not in proxy.transport:
            base['uuid'] = f'{hconfigs[ConfigEnum.shared_secret]}'

    if base["proto"] == "ssr":
        base["ssr-obfs"] = "tls1.2_ticket_auth"
        base["ssr-protocol"] = "auth_sha1_v4"
        base["fakedomain"] = hconfigs[ConfigEnum.ssr_fakedomain]
        base["mode"] = "FakeTLS"
        return base
    elif "faketls" in proxy.transport:
        base['fakedomain'] = hconfigs[ConfigEnum.ssfaketls_fakedomain]
        base['mode'] = 'FakeTLS'
        return base
    elif "shadowtls" in proxy.transport:

        base['fakedomain'] = hconfigs[ConfigEnum.shadowtls_fakedomain]
        base['mode'] = 'ShadowTLS'
        return base

    if ProxyTransport.XTLS in proxy.transport:
        base['flow'] = 'xtls-rprx-vision'
        return {**base, 'transport': 'tcp'}
    if "tcp" in proxy.transport:
        base['transport'] = 'tcp'
        base['path'] = f'/{path[base["proto"]]}{hconfigs[ConfigEnum.path_tcp]}'
        return base
    if proxy.transport in ["ws", "WS"]:
        base['transport'] = 'ws'
        base['path'] = f'/{path[base["proto"]]}{hconfigs[ConfigEnum.path_ws]}'
        base["host"] = domain
        return base

    if proxy.transport == "grpc":
        base['transport'] = 'grpc'
        # base['grpc_mode'] = "multi" if hconfigs[ConfigEnum.core_type]=='xray' else 'gun'
        base['grpc_mode'] = 'gun'
        base['grpc_service_name'] = f'{path[base["proto"]]}{hconfigs[ConfigEnum.path_grpc]}'
        base['path'] = base['grpc_service_name']
        return base

    if "h1" in proxy.transport:
        base['transport'] = 'tcp'
        base['alpn'] = 'http/1.1'
        return base
    if ProxyProto.ssh == proxy.proto:
        base['private_key'] = g.user.ed25519_private_key
        base['host_key'] = hiddify.get_hostkeys(False)
        # base['ssh_port'] = hconfig(ConfigEnum.ssh_server_port)
        return base
    return {'name': name, 'msg': 'not valid', 'type': 'error', 'proto': proxy.proto}


def to_link(proxy):
    if 'error' in proxy:
        return proxy
    orig_name_link = (proxy['extra_info'] + " " + proxy["name"]).strip()
    name_link = hiddify.url_encode(orig_name_link)
    if proxy['proto'] == 'vmess':
        # print(proxy)
        vmess_type = None
        if proxy["transport"] == 'tcp':
            vmess_type = 'http'
        if 'grpc_mode' in proxy:
            vmess_type = proxy['grpc_mode']
        vmess_data = {"v": "2",
                      "ps": orig_name_link,
                      "add": proxy['server'],
                      "port": proxy['port'],
                      "id": proxy["uuid"],
                      "aid": 0,
                      "scy": proxy['cipher'],
                      "net": proxy["transport"],
                      "type": vmess_type or "none",
                      "host": proxy.get("host", ""),
                      "alpn": proxy.get("alpn", "h2,http/1.1"),
                      "path": proxy["path"] if "path" in proxy else "",
                      "tls": "tls" if "tls" in proxy["l3"] else "",
                      "sni": proxy["sni"],
                      "fp": proxy["fingerprint"]
                      }
        if 'reality' in proxy["l3"]:
            vmess_data['tls'] = "reality"
            vmess_data['pbk'] = proxy['reality_pbk']
            vmess_data['sid'] = proxy['reality_short_id']

        return "vmess://" + hiddify.do_base_64(f'{json.dumps(vmess_data)}')
        # return pbase64(f'vmess://{json.dumps(vmess_data)}')
    if proxy['proto'] == 'ssh':
        strenssh = hiddify.do_base_64(f'{proxy["uuid"]}:0:{proxy["private_key"]}::@{proxy["server"]}:{proxy["port"]}')
        baseurl = f'ssh://{strenssh}#{name_link}'
        baseurl += f'\nssh://{proxy["uuid"]}@{proxy["server"]}:{proxy["port"]}/?pk='+hiddify.do_base_64(proxy["private_key"])+f"&hk={hiddify.get_hostkeys(True)}#{name_link}"

        return baseurl
    if proxy['proto'] == "ssr":
        baseurl = f'ssr://{proxy["cipher"]}:{proxy["uuid"]}@{proxy["server"]}:{proxy["port"]}'
        return baseurl
    if proxy['proto'] in ['ss', 'v2ray']:
        baseurl = f'ss://{proxy["cipher"]}:{proxy["uuid"]}@{proxy["server"]}:{proxy["port"]}'
        if proxy['mode'] == 'faketls':
            return f'{baseurl}?plugin=obfs-local%3Bobfs%3Dtls%3Bobfs-host%3D{proxy["fakedomain"]}%3Budp-over-tcp=true#{name_link}'
        # if proxy['mode'] == 'shadowtls':
        #     return f'{baseurl}?plugin=shadow-tls%3Bpassword%3D{proxy["proxy_path"]}%3Bhost%3D{proxy["fakedomain"]}%3Budp-over-tcp=true#{name_link}'
        if proxy['proto'] == 'v2ray':
            return f'{baseurl}?plugin=v2ray-plugin%3Bmode%3Dwebsocket%3Bpath%3D{proxy["path"]}%3Bhost%3D{proxy["host"]}%3Btls%3Budp-over-tcp=true#{name_link}'
    baseurl = f'{proxy["proto"]}://{proxy["uuid"]}@{proxy["server"]}:{proxy["port"]}?hiddify=1'
    baseurl += f'&sni={proxy["sni"]}&type={proxy["transport"]}'

    baseurl += f"&alpn={proxy['alpn']}"

    # infos+=f'&alpn={proxy["alpn"]}'
    baseurl += f'&path={proxy["path"]}' if "path" in proxy else ""
    baseurl += f'&host={proxy["host"]}' if "host" in proxy else ""
    if "grpc" == proxy["transport"]:
        baseurl += f'&serviceName={proxy["grpc_service_name"]}&mode={proxy["grpc_mode"]}'
    # print(proxy['cdn'],proxy["transport"])
    if request.args.get("fragment"):
        baseurl += f'&fragment='+request.args.get("fragment")
    if "ws" == proxy["transport"] and proxy['cdn'] and request.args.get("fragment_v1"):
        baseurl += f'&fragment_v1='+request.args.get("fragment_v1")
    if 'vless' == proxy['proto']:
        baseurl += "&encryption=none"
    if proxy.get('fingerprint', 'none') != 'none':
        baseurl += "&fp="+proxy['fingerprint']
    if proxy['l3'] != 'quic':
        baseurl += '&headerType=None'  # if not quic
    if proxy['mode'] == 'Fake' or proxy['allow_insecure']:
        baseurl += "&allowInsecure=true"
    if proxy.get('flow'):
        baseurl += f'&flow={proxy["flow"]}'

    infos = f'#{name_link}'

    if 'reality' in proxy["l3"]:
        return f"{baseurl}&security=reality&pbk={proxy['reality_pbk']}&sid={proxy['reality_short_id']}{infos}"
    if 'tls' in proxy['l3']:
        return f'{baseurl}&security=tls{infos}'
    if proxy['l3'] == 'http':
        return f'{baseurl}&security=none{infos}'
    return proxy


def to_clash_yml(proxy):
    return yaml.dump(to_clash(proxy))


def to_clash(proxy, meta_or_normal):

    name = proxy['name']
    if proxy['l3'] == "kcp":
        return {'name': name, 'msg': "clash does not support kcp", 'type': 'debug'}
    if proxy['proto'] == "ssh":
        return {'name': name, 'msg': "clash does not support ssh", 'type': 'debug'}
    if meta_or_normal == "normal":
        if proxy['proto'] == "vless":
            return {'name': name, 'msg': "vless not supported in clash", 'type': 'debug'}
        if proxy.get('flow'):
            return {'name': name, 'msg': "xtls not supported in clash", 'type': 'debug'}
        if proxy['transport'] == "shadowtls":
            return {'name': name, 'msg': "shadowtls not supported in clash", 'type': 'debug'}
    if proxy['l3'] == ProxyL3.tls_h2 and proxy['proto'] in [ProxyProto.vmess, ProxyProto.vless] and proxy['dbe'].cdn == ProxyCDN.direct:
        return {'name': name, 'msg': "bug tls_h2 vmess and vless in clash meta", 'type': 'warning'}
    base = {}
    # vmess ws
    base["name"] = f"""{proxy['extra_info']} {proxy["name"]} {proxy['port']} {proxy["dbdomain"].id}"""
    base["type"] = str(proxy["proto"])
    base["server"] = proxy["server"]
    base["port"] = proxy["port"]
    base['alpn'] = proxy['alpn'].split(',')
    if proxy["proto"] == "ssr":
        base["cipher"] = proxy["cipher"]
        base["password"] = proxy["uuid"]
        base["udp"] = True
        base["obfs"] = proxy["ssr-obfs"]
        base["protocol"] = proxy["ssr-protocol"]
        base["obfs-param"] = proxy["fakedomain"]
        return base
    elif proxy["proto"] in ["ss", "v2ray"]:
        base["cipher"] = proxy["cipher"]
        base["password"] = proxy["uuid"]
        base["udp_over_tcp"] = True
        if proxy["transport"] == "faketls":
            base["plugin"] = "obfs"
            base["plugin-opts"] = {
                "mode": 'tls',
                "host": proxy["fakedomain"]
            }
        elif proxy["transport"] == "shadowtls":
            base["plugin"] = "shadow-tls"
            base["plugin-opts"] = {
                "host": proxy["fakedomain"],
                "password": proxy["proxy_path"],
                "version": 3  # support 1/2/3

            }

        elif proxy["proto"] == "v2ray":
            base["plugin"] = "v2ray-plugin"
            base["type"] = "ss"
            base["plugin-opts"] = {
                "mode": "websocket",
                "tls": "tls" in proxy["l3"],
                "skip-cert-verify": proxy["mode"] == "Fake" or proxy['allow_insecure'],
                "host": proxy['sni'],
                "path": proxy["path"]
            }
        return base
    elif proxy["proto"] == "trojan":
        base["password"] = proxy["uuid"]
        base["sni"] = proxy["sni"]

    else:
        base["uuid"] = proxy["uuid"]
        base["servername"] = proxy["sni"]
        base["tls"] = "tls" in proxy["l3"] or "reality" in proxy["l3"]
    if meta_or_normal == "meta":
        base['client-fingerprint'] = proxy['fingerprint']
    if proxy.get('flow'):
        base["flow"] = proxy['flow']
        # base["flow-show"] = True

    if proxy["proto"] == "vmess":
        base["alterId"] = 0
        base["cipher"] = proxy["cipher"]
    base["udp"] = True

    base["skip-cert-verify"] = proxy["mode"] == "Fake"

    base["network"] = proxy["transport"]

    if base["network"] == "ws":
        base["ws-opts"] = {
            "path": proxy["path"]
        }
        if "host" in proxy:
            base["ws-opts"]["headers"] = {"Host": proxy["host"]}

    if base["network"] == "tcp" and proxy['alpn'] != 'h2':
        if proxy['transport'] != ProxyTransport.XTLS:
            base["network"] = "http"

        if "path" in proxy:
            base["http-opts"] = {
                "path": [proxy["path"]]
            }
            if 'host' in proxy:
                base["http-opts"]["host"] = [proxy["host"]]
    if base["network"] == "tcp" and proxy['alpn'] == 'h2':
        base["network"] = "h2"

        if "path" in proxy:
            base["h2-opts"] = {
                "path": proxy["path"]
            }
            if 'host' in proxy:
                base["h2-opts"]["host"] = [proxy["host"]]
    if base["network"] == "grpc":
        base["grpc-opts"] = {
            "grpc-service-name": proxy["grpc_service_name"]
        }
    if proxy['l3'] == ProxyL3.reality:
        base["reality-opts"] = {
            "public-key": proxy['reality_pbk'],
            "short-id": proxy['reality_short_id'],
        }
        if proxy["transport"] != 'grpc':
            base["network"] = 'tcp'

    return base


def get_clash_config_names(meta_or_normal, domains):
    allp = []
    for pinfo in get_all_validated_proxies(domains):
        clash = to_clash(pinfo, meta_or_normal)
        if 'msg' not in clash:
            allp.append(clash['name'])

    return yaml.dump(allp, sort_keys=False)


def get_all_clash_configs(meta_or_normal, domains):
    allp = []
    for pinfo in get_all_validated_proxies(domains):
        clash = to_clash(pinfo, meta_or_normal)
        if 'msg' not in clash:
            allp.append(clash)

    return yaml.dump({"proxies": allp}, sort_keys=False)


def to_singbox(proxy):
    name = proxy['name']

    all_base = []
    if proxy['l3'] == "kcp":
        return {'name': name, 'msg': "clash does not support kcp", 'type': 'debug'}

    base = {}
    all_base.append(base)
    # vmess ws
    base["tag"] = f"""{proxy['extra_info']} {proxy["name"]} {proxy['port']} {proxy["dbdomain"].id}"""
    base["type"] = str(proxy["proto"])
    base["server"] = proxy["server"]
    base["server_port"] = int(proxy["port"])
    # base['alpn'] = proxy['alpn'].split(',')
    if proxy["proto"] == "ssr":
        add_singbox_ssr(base, proxy)
        return all_base
    if proxy["proto"] in ["ss", "v2ray"]:
        add_singbox_shadowsocks_base(all_base, proxy)
        return all_base
    if proxy["proto"] == "ssh":
        add_singbox_ssh(all_base, proxy)
        return all_base
    if proxy["proto"] == "trojan":
        base["password"] = proxy["uuid"]
    elif proxy['proto'] in ['trojan', 'vmess', 'vless']:
        base["uuid"] = proxy["uuid"]
        add_singbox_multiplex(base)

    add_singbox_tls(base, proxy)

    if proxy.get('flow'):
        base["flow"] = proxy['flow']
        # base["flow-show"] = True

    if proxy["proto"] == "vmess":
        base["alter_id"] = 0
        base["security"] = proxy["cipher"]

    # base["udp"] = True
    if proxy["proto"] in ["vmess", "vless"]:
        base["packet_encoding"] = "xudp"  # udp packet encoding
    add_singbox_transport(base, proxy)

    return all_base


def add_singbox_multiplex(base):
    return
    base['multiplex'] = {
        "enabled": True,
        "protocol": "h2mux",
        "max_connections": 4,
        "min_streams": 4,
        "max_streams": 0,
        "padding": false
    }


def add_singbox_udp_over_tcp(base):
    base['udp_over_tcp'] = {
        "enabled": True,
        "version": 2
    }


def add_singbox_tls(base, proxy):
    if not ("tls" in proxy["l3"] or "reality" in proxy["l3"]):
        return
    base["tls"] = {
        "enabled": True,
        "server_name": proxy["sni"],
        "utls": {
            "enabled": True,
            "fingerprint": proxy.get('fingerprint', 'none')
        }
    }

    if "reality" in proxy["l3"]:
        base["tls"]["reality"] = {
            "enabled": True,
            "public_key": proxy['reality_pbk'],
            "short_id": proxy['reality_short_id']
        }
    base["tls"]['insecure'] = proxy['allow_insecure'] or (proxy["mode"] == "Fake")
    base["tls"]["alpn"] = proxy['alpn'].split(',')
    # base['ech'] = {
    #     "enabled": True,
    # }


def add_singbox_transport(base, proxy):
    if proxy['l3'] == 'reality':
        return
    base["transport"] = {}
    if proxy['transport'] in ["ws", "WS"]:
        base["transport"] = {
            "type": "ws",
            "path": proxy["path"],
            "early_data_header_name": "Sec-WebSocket-Protocol"
        }
        if "host" in proxy:
            base["transport"]["headers"] = {"Host": proxy["host"]}

    if proxy["transport"] == "tcp":
        base["transport"] = {
            "type": "http",
            "path": proxy.get("path", ""),
            # "method": "",
            # "headers": {},
            "idle_timeout": "15s",
            "ping_timeout": "15s"
        }

        if 'host' in proxy:
            base["transport"]["host"] = [proxy["host"]]

    if proxy["transport"] == "grpc":
        base["transport"] = {
            "type": "grpc",
            "service_name": proxy["grpc_service_name"],
            "idle_timeout": "115s",
            "ping_timeout": "15s",
            # "permit_without_stream": false
        }


def add_singbox_ssr(base, proxy):

    base["method"] = proxy["cipher"]
    base["password"] = proxy["uuid"]
    # base["udp"] = True
    base["obfs"] = proxy["ssr-obfs"]
    base["protocol"] = proxy["ssr-protocol"]
    base["protocol-param"] = proxy["fakedomain"]


def add_singbox_shadowsocks_base(all_base, proxy):
    base = all_base[0]
    base["type"] = "shadowsocks"
    base["method"] = proxy["cipher"]
    base["password"] = proxy["uuid"]
    add_singbox_udp_over_tcp(base)
    add_singbox_multiplex(base)
    if proxy["transport"] == "faketls":
        base["plugin"] = "obfs-local"
        base["plugin_opts"] = f'obfs=tls&obfs-host={proxy["fakedomain"]}'
    if proxy['proto'] == 'v2ray':
        base["plugin"] = "v2ray-plugin"
        # "skip-cert-verify": proxy["mode"] == "Fake" or proxy['allow_insecure'],
        base["plugin_opts"] = f'mode=websocket&path={proxy["path"]}&host={proxy["host"]}&tls'

    if proxy["transport"] == "shadowtls":
        base['detour'] = base['tag']+"_shadowtls-out"

        shadowtls_base = {
            "type": "shadowtls",
            "tag": base['tag']+"_shadowtls-out",
            "server": base['server'],
            "server_port": base['server_port'],
            "version": 3,
            "password": proxy["proxy_path"],
            "tls": {
                "enabled": True,
                "server_name": proxy["fakedomain"],
                "utls": {
                    "enabled": True,
                    "fingerprint": proxy.get('fingerprint', 'none')
                }
            }
        }
        # add_singbox_utls(shadowtls_base)
        del base['server']
        del base['server_port']
        all_base.append(shadowtls_base)


def add_singbox_ssh(all_base, proxy):
    base = all_base[0]
    # base["client_version"]= "{{ssh_client_version}}"
    base["user"] = proxy['uuid']
    base["private_key"] = proxy['private_key']  # .replace('\n', '\\n')

    base["host_key"] = proxy.get('host_key', [])

    socks_front = {
        "type": "socks",
        "tag": base['tag']+"+UDP",
        "server": "127.0.0.1",
        "server_port": 2000,
        "version": "5",
        "udp_over_tcp": True,
        "network": "tcp",
        "detour": base['tag']
    }
    all_base.append(socks_front)


def make_full_singbox_config(domains, **kwargs):
    ua = hiddify.get_user_agent()
    base_config = json.loads(render_template('base_singbox_config.json'))
    allphttp = [p for p in request.args.get("phttp", "").split(',') if p]
    allptls = [p for p in request.args.get("ptls", "").split(',') if p]

    allp = []
    for d in domains:
        base_config['dns']['rules'][0]['domain'].append(d.domain)
    for pinfo in get_all_validated_proxies(domains):
        sing = to_singbox(pinfo)
        if 'msg' not in sing:
            allp += sing
    base_config['outbounds'] += allp

    select = {
        "type": "selector",
        "tag": "Select",
        "outbounds": [p['tag'] for p in allp if 'shadowtls-out' not in p['tag']],
        "default": "Auto"
    }
    select['outbounds'].insert(0, "Auto")
    base_config['outbounds'].insert(0, select)
    smart = {
        "type": "urltest",
        "tag": "Auto",
        "outbounds": [p['tag'] for p in allp if 'shadowtls-out' not in p],
        "url": "https://www.gstatic.com/generate_204",
        "interval": "10m",
        "tolerance": 200
    }
    base_config['outbounds'].insert(1, smart)
    res = json.dumps(base_config, indent=4)
    if ua['is_hiddify']:
        res = res[:-1]+',"experimental": {}}'
    return res


def make_v2ray_configs(user, user_activate, domains, expire_days, ip_debug, db_domain, has_auto_cdn, asn, profile_title, **kwargs):
    res = []

    ua = hiddify.get_user_agent()
    if hconfig(ConfigEnum.show_usage_in_sublink):

        if not ua['is_hiddify']:

            fake_ip_for_sub_link = datetime.datetime.now().strftime(f"%H.%M--%Y.%m.%d.time:%H%M")
            # if ua['app'] == "Fair1":
            #     res.append(f'trojan://1@{fake_ip_for_sub_link}?sni=fake_ip_for_sub_link&security=tls#{round(user.current_usage_GB,3)}/{user.usage_limit_GB}GB_Remain:{expire_days}days')
            # else:

            # res.append(f'trojan://1@{fake_ip_for_sub_link}?sni=fake_ip_for_sub_link&security=tls#{hiddify.url_encode(profile_title)}')

            name = '⏳' if user_activate else '✖'
            if user.usage_limit_GB < 1000:
                name += f'{round(user.current_usage_GB,3)}/{user.usage_limit_GB}GB'
            elif user.usage_limit_GB < 100000:
                name += f'{round(user.current_usage_GB/1000,3)}/{round(user.usage_limit_GB/1000,1)}TB'
            else:
                res.append("#Unlimited usage")
            if expire_days < 1000:
                name += " "+_(f'📅%(expire_days)s days', expire_days=expire_days)
            else:
                res.append("#Unlimited Time")

            name = name.strip()
            if len(name) > 3:
                res.append(f'trojan://1@{fake_ip_for_sub_link}?sni=fake_ip_for_sub_link&security=tls#{hiddify.url_encode(name)}')

    if ua['is_browser']:
        res.append(f'#Hiddify auto ip: {ip_debug}')

    if not user_activate:

        if hconfig(ConfigEnum.lang) == 'fa':
            res.append('trojan://1@1.1.1.1#'+hiddify.url_encode('✖بسته شما به پایان رسید'))
        else:
            res.append('trojan://1@1.1.1.1#'+hiddify.url_encode('✖Package_Ended'))
        return "\n".join(res)

    for pinfo in get_all_validated_proxies(domains):
        link = to_link(pinfo)
        if 'msg' not in link:
            res.append(link)
    return "\n".join(res)


def get_all_validated_proxies(domains):
    allp = []
    allphttp = [p for p in request.args.get("phttp", "").split(',') if p]
    allptls = [p for p in request.args.get("ptls", "").split(',') if p]
    added_ip = {'ssh': {}, 'tuic': {}, 'hysteria': {}}
    for d in domains:
        # raise Exception(base_config)
        hconfigs = get_hconfigs(d.child_id)
        for type in all_proxies(d.child_id):
            options = []
            if type.proto in ['ssh', 'tuic', 'hysteria']:

                ip = hiddify.get_domain_ip(d.domain, version=4)
                ip6 = hiddify.get_domain_ip(d.domain, version=6)

                ips = [x for x in [ip, ip6]if x != None]

                if all([x in added_ip[type.proto] for x in ips]):
                    print('skiping ')
                    continue
                for x in ips:
                    added_ip[type.proto][x] = 1

                if type.proto == 'ssh':
                    if d.mode == 'fake':
                        continue
                    options = [{'pport': hconfigs[ConfigEnum.ssh_server_port]}]
                elif type.proto == 'tuic':
                    options = [{'pport': hconfigs[ConfigEnum.tuic_port]}]
                elif type.proto == 'hysteria':
                    options = [{'pport': hconfigs[ConfigEnum.hysteria_port]}]
            else:
                for t in (['http', 'tls'] if hconfigs[ConfigEnum.http_proxy_enable] else ['tls']):
                    for port in hconfigs[ConfigEnum.http_ports if t == 'http' else ConfigEnum.tls_ports].split(','):
                        phttp = port if t == 'http' else None
                        ptls = port if t == 'tls' else None
                        if phttp and len(allphttp) and phttp not in allphttp:
                            continue
                        if ptls and len(allptls) and ptls not in allptls:
                            continue
                        options.append({'phttp': phttp, 'ptls': ptls})

            for opt in options:
                pinfo = make_proxy(type, d, **opt)
                if 'msg' not in pinfo:
                    allp.append(pinfo)
    return allp



from hiddifypanel.models import *
import flask_babel
import flask_babelex

# from flask_babelex import lazy_gettext as _
from flask_babelex import gettext as _
import wtforms as wtf
from flask_wtf import FlaskForm
from flask_bootstrap import SwitchField
from hiddifypanel.panel import hiddify
from flask_admin.base import expose
# from gettext import gettext as _
from wtforms.validators import Regexp, ValidationError

import re
from flask import render_template, current_app, Markup, redirect, url_for
from hiddifypanel.models import User, Domain, DomainType, StrConfig, ConfigEnum, get_hconfigs
from hiddifypanel.panel.database import db
from wtforms.fields import *

from hiddifypanel.panel.hiddify import flash
from flask_classful import FlaskView


class QuickSetup(FlaskView):

    def index(self):
        return render_template('quick_setup.html', lang_form=get_lang_form(), form=get_quick_setup_form(), ipv4=hiddify.get_ip(4), ipv6=hiddify.get_ip(6), admin_link=admin_link(), show_domain_info=True)

    def post(self):
        set_hconfig(ConfigEnum.first_setup, False)
        quick_form = get_quick_setup_form()
        lang_form = get_lang_form()
        if lang_form.lang_submit.data:
            if lang_form.validate_on_submit():
                StrConfig.query.filter(StrConfig.key == ConfigEnum.lang, StrConfig.child_id == 0).first().value = lang_form.admin_lang.data
                StrConfig.query.filter(StrConfig.key == ConfigEnum.admin_lang, StrConfig.child_id == 0).first().value = lang_form.admin_lang.data
                StrConfig.query.filter(StrConfig.key == ConfigEnum.country, StrConfig.child_id == 0).first().value = lang_form.country.data
                db.session.commit()

                flask_babel.refresh()
                flask_babelex.refresh()
                # with flask_babel.force_locale(lang_form.admin_lang.data):
                #         flask_babel.refresh()
                #         flask_babelex.refresh()
                flash((_('quicksetup.setlang.success')), 'success')
            else:
                flash((_('quicksetup.setlang.error')), 'danger')

            return render_template('quick_setup.html', form=get_quick_setup_form(True), lang_form=get_lang_form(), admin_link=admin_link(), ipv4=hiddify.get_ip(4), ipv6=hiddify.get_ip(6), show_domain_info=False)

        if quick_form.validate_on_submit():
            sslip_dm = Domain.query.filter(Domain.domain == f'{hiddify.get_ip(4)}.sslip.io').delete()
            db.session.add(Domain(domain=quick_form.domain.data.lower(), mode=DomainType.direct))
            hiddify.bulk_register_configs([
                {"key": ConfigEnum.telegram_enable, "value": quick_form.enable_telegram.data == True},
                {"key": ConfigEnum.vmess_enable, "value": quick_form.enable_vmess.data == True},
                {"key": ConfigEnum.firewall, "value": quick_form.enable_firewall.data == True},
                {"key": ConfigEnum.block_iran_sites, "value": quick_form.block_iran_sites.data == True},
                # {"key":ConfigEnum.decoy_domain,"value":quick_form.decoy_domain.data}
            ])

            db.session.commit()
            # hiddify.flash_config_success()
            proxy_path = hconfig(ConfigEnum.proxy_path)
            # uuid=User.query.first().uuid
            # userlink=f"<a class='btn btn-secondary share-link' target='_blank' href='https://{quick_form.domain.data}/{proxy_path}/{uuid}/'>{_('default user link')}</a>"
            # flash((_('The default user link is %(link)s. To add or edit more users, please visit users from menu.',link=userlink)),'info')

            from . import Actions
            action = Actions()
            return action.reinstall(domain_changed=True)
        else:
            flash(_('config.validation-error'), 'danger')
        return render_template('quick_setup.html', form=quick_form, lang_form=get_lang_form(True), ipv4=hiddify.get_ip(4), ipv6=hiddify.get_ip(6), admin_link=admin_link(), show_domain_info=False)


def get_lang_form(empty=False):
    class LangForm(FlaskForm):
        admin_lang = wtf.fields.SelectField(_("config.admin_lang.label"), choices=[("en", _("lang.en")), ("fa", _("lang.fa")), ("pt", _(
            "lang.pt")), ("zh", _("lang.zh"))], description=_("config.admin_lang.description"), default=hconfig(ConfigEnum.admin_lang))
        # lang=wtf.fields.SelectField(_("config.lang.label"),choices=[("en",_("lang.en")),("fa",_("lang.fa"))],description=_("config.lang.description"),default=hconfig(ConfigEnum.lang))
        country = wtf.fields.SelectField(_("config.country.label"), choices=[("ir", _("Iran")), ("zh", _(
            "China")), ("other", "Others")], description=_("config.country.description"), default=hconfig(ConfigEnum.country))
        lang_submit = wtf.fields.SubmitField(_('Submit'))

    return LangForm(None)if empty else LangForm()


def get_quick_setup_form(empty=False):
    def get_used_domains():
        configs = get_hconfigs()
        domains = []
        for c in configs:
            if "domain" in c:
                domains.append(configs[c])
        for d in Domain.query.all():
            domains.append(d.domain)
        return domains

    class QuickSetupForm(FlaskForm):
        domain_regex = "^([A-Za-z0-9\-\.]+\.[a-zA-Z]{2,})$"

        domain_validators = [wtf.validators.Regexp(domain_regex, re.IGNORECASE, _("config.Invalid domain")),
                             validate_domain,
                             wtf.validators.NoneOf([d.domain.lower() for d in Domain.query.all()], _("config.Domain already used")),
                             wtf.validators.NoneOf([c.value.lower() for c in StrConfig.query.all() if "fakedomain" in c.key and c.key !=
                                                   ConfigEnum.decoy_domain], _("config.Domain already used"))
                             ]
        domain = wtf.fields.StringField(_("domain.domain"), domain_validators, description=_("domain.description"), render_kw={
                                        "class": "ltr", "pattern": domain_validators[0].regex.pattern, "title": domain_validators[0].message, "required": "", "placeholder": "sub.domain.com"})
        enable_telegram = SwitchField(_("config.telegram_enable.label"), description=_("config.telegram_enable.description"), default=hconfig(ConfigEnum.telegram_enable))
        enable_firewall = SwitchField(_("config.firewall.label"), description=_("config.firewall.description"), default=hconfig(ConfigEnum.firewall))
        block_iran_sites = SwitchField(_("config.block_iran_sites.label"), description=_("config.block_iran_sites.description"), default=hconfig(ConfigEnum.block_iran_sites))
        enable_vmess = SwitchField(_("config.vmess_enable.label"), description=_("config.vmess_enable.description"), default=hconfig(ConfigEnum.vmess_enable))
        decoy_domain=wtf.fields.StringField(_("config.decoy_domain.label"),description=_("config.decoy_domain.description"),default=hconfig(ConfigEnum.decoy_domain),validators=[wtf.validators.Regexp(domain_regex,re.IGNORECASE,_("config.Invalid domain")),hiddify.validate_domain_exist])
        submit = wtf.fields.SubmitField(_('Submit'))

    return QuickSetupForm(None) if empty else QuickSetupForm()


def validate_domain(form, field):
    domain = field.data
    dip = hiddify.get_domain_ip(domain)
    if dip == None:
        raise ValidationError(_("Domain can not be resolved! there is a problem in your domain"))

    myip = hiddify.get_ip(4)
    if dip and myip != dip:
        raise ValidationError(_("Domain (%(domain)s)-> IP=%(domain_ip)s is not matched with your ip=%(server_ip)s which is required in direct mode",
                              server_ip=myip, domain_ip=dip, domain=domain))


def admin_link():
    domains = get_panel_domains()
    if len(domains):
        return "https://"+domains[0].domain+hiddify.get_admin_path()
    else:
        return "https://"+hiddify.get_ip(4)+hiddify.get_admin_path()




from sqlalchemy.orm import Load, joinedload
from .ssh_liberty_bridge_api import SSHLibertyBridgeApi
from .xray_api import XrayApi
from .singbox_api import SingboxApi
from hiddifypanel.models import *
from hiddifypanel.panel import hiddify
drivers = [XrayApi(), SingboxApi(), SSHLibertyBridgeApi()]


def get_users_usage(reset=True):
    res = {}
    for user in User.query.all():
        d = 0
        for driver in drivers:
            try:
                d += driver.get_usage(user.uuid, reset=True) or 0
            except Exception as e:
                hiddify.error(f'ERROR! {driver.__class__.__name__} has error {e} in get_usage for user={user.uuid}')
        res[user] = {'usage': d, 'ips': ''}
    return res


def get_enabled_users():
    from collections import defaultdict
    d = defaultdict(int)
    total = 0
    for driver in drivers:
        try:
            for u, v in driver.get_enabled_users().items():
                if not v:
                    continue
                d[u] += 1
            total += 1
        except Exception as e:
            hiddify.error(f'ERROR! {driver.__class__.__name__} has error in get_enabled users')

    res = defaultdict(bool)
    for u, v in d.items():
        res[u] = v >= total  # ignore singbox
    return res


def add_client(user):
    for driver in drivers:
        try:
            driver.add_client(user)
        except Exception as e:
            hiddify.error(f'ERROR! {driver.__class__.__name__} has error {e} in add client for user={user.uuid}')


def remove_client(user):
    for driver in drivers:
        try:
            driver.remove_client(user)
        except Exception as e:
            hiddify.error(f'ERROR! {driver.__class__.__name__} has error {e} in remove client for user={user.uuid}')



from flask import Blueprint, url_for, request, jsonify, g, redirect
from flask_admin import Admin
from hiddifypanel import Events
# from flask_sockets import Sockets

from .SettingAdmin import SettingAdmin
from .DomainAdmin import DomainAdmin
from .ConfigAdmin import ConfigAdmin
from .commercial_info import CommercialInfo
from .ProxyAdmin import ProxyAdmin
from .AdminstratorAdmin import AdminstratorAdmin
from .Actions import Actions
from .Backup import Backup
from .QuickSetup import QuickSetup
from hiddifypanel.models import StrConfig, ConfigEnum
# from .resources import ProductItemResource, ProductResource
from hiddifypanel.panel.database import db
from hiddifypanel.models import *
from .Dashboard import Dashboard

from flask_admin.menu import MenuLink

import uuid
from flask_adminlte3 import AdminLTE3
flask_bp = Blueprint("flask", __name__, url_prefix=f"/<proxy_path>/<user_secret>/", template_folder="templates")
admin_bp = Blueprint("admin", __name__, url_prefix=f"/<proxy_path>/<user_secret>/admin/", template_folder="templates")
flaskadmin = Admin(endpoint="admin", base_template='flaskadmin-layout.html')
# from extensions import socketio


def init_app(app):

    from .UserAdmin import UserAdmin
    # admin_secret=StrConfig.query.filter(StrConfig.key==ConfigEnum.admin_secret).first()
    #
    # return
    # admin = Admin(endpoint="admin",index_view=Dashboard(),base_template='lte-master.html',static_url_path="/static/")
    flaskadmin.template_mode = "bootstrap4"
    flaskadmin.init_app(flask_bp)
    adminlte = AdminLTE3()
    adminlte.init_app(app)

    Events.admin_prehook.notify(flaskadmin=flaskadmin, admin_bp=admin_bp)

    @app.route('/<proxy_path>/<user_secret>/admin')
    def auto_route():
        return redirect(request.url.replace("http://", "https://")+"/")

    # flaskadmin.init_app(app)

    flaskadmin.add_view(UserAdmin(User, db.session))
    flaskadmin.add_view(DomainAdmin(Domain, db.session))
    flaskadmin.add_view(AdminstratorAdmin(AdminUser, db.session))

    SettingAdmin.register(admin_bp)
    ProxyAdmin.register(admin_bp)
    Actions.register(admin_bp)
    CommercialInfo.register(admin_bp)

    QuickSetup.register(admin_bp)
    Backup.register(admin_bp)
    Dashboard.register(admin_bp, route_base="/")

    # admin_bp.add_url_rule('/admin/quicksetup/',endpoint="quicksetup",view_func=QuickSetup.index,methods=["GET"])
    # admin_bp.add_url_rule('/admin/quicksetup/',endpoint="quicksetup-save", view_func=QuickSetup.save,methods=["POST"])

    app.add_url_rule("/<proxy_path>/<user_secret>/admin/static/<filename>/", endpoint="admin.static")  # fix bug in admin with blueprint

    flask_bp.debug = True
    app.register_blueprint(admin_bp)
    app.register_blueprint(flask_bp)



from flask import request, flash
import random
import re
import os
import maxminddb
from hiddifypanel.models import *
import sys
from flask_babelex import gettext as _
default_ips = """
mci.ircf.space		MCI
mcix.ircf.space		MCI
mtn.ircf.space		MTN
mtnx.ircf.space		MTN
mkh.ircf.space		MKH
mkhx.ircf.space		MKH
rtl.ircf.space		RTL
hwb.ircf.space		HWB
ast.ircf.space		AST
sht.ircf.space		SHT
prs.ircf.space		PRS
mbt.ircf.space		MBT
ask.ircf.space		ASK
rsp.ircf.space		RSP
afn.ircf.space		AFN
ztl.ircf.space		ZTL
psm.ircf.space		PSM
arx.ircf.space		ARX
smt.ircf.space		SMT
fnv.ircf.space		FNV
dbn.ircf.space		DBN
apt.ircf.space		APT
"""

try:
    ipasn = maxminddb.open_database('GeoLite2-ASN.mmdb') if os.path.exists('GeoLite2-ASN.mmdb') else {}
    ipcountry = maxminddb.open_database('GeoLite2-Country.mmdb') if os.path.exists('GeoLite2-Country.mmdb') else {}
    ipcity = maxminddb.open_database('GeoLite2-City.mmdb') if os.path.exists('GeoLite2-City.mmdb') else {}
except Exception as e:
    print("Error can not load maxminddb", file=sys.stderr)
    ipasn = {}
    ipcountry = {}

asn_map = {
    '58224': 'MKH',
    '197207': 'MCI',
    '12880': 'ITC',
    '44244': 'MTN',
    '57218': 'RTL',
    '16322': 'PRS',
    '56402': 'HWB',
    '41689': 'AST',
    '43754': 'AST',
    '31549': 'SHT',
    '205647': 'SHT',
    '50810': 'MBT',
    '39308': 'ASK',
    '205207': 'RSP',
    '25184': 'AFR',
    '394510': 'ZTL',
    '206065': 'ZTL',
    '49100': 'PSM'
}


def get_asn_short_name(user_ip=None):
    user_ip = user_ip or get_real_user_ip()
    try:
        asn_id = get_asn_id(user_ip)
        return asn_map.get(asn_id, "unknown")
    except:
        return "unknown"


def get_asn_id(user_ip=None):
    user_ip = user_ip or get_real_user_ip()
    try:
        asnres = ipasn.get(user_ip)
        return asnres['autonomous_system_number']
    except:
        return "unknown"


def get_country(user_ip=None):
    try:
        user_ip = user_ip or get_real_user_ip()
        return (ipcountry.get(user_ip) or {}).get('country', {}).get('iso_code', 'unknown')
    except:
        return 'unknown'


def get_city(user_ip=None):
    try:
        user_ip = user_ip or get_real_user_ip()
        res = ipcity.get(user_ip)
        return {'city': res.get('city').get('name'), 'latitude': res.get('latitude'), 'longitude': res.get('longitude'), 'accuracy_radius': res.get('accuracy_radius')}
    except:
        return 'unknown'


def get_real_user_ip_debug(user_ip=None):
    user_ip = user_ip or get_real_user_ip()
    asnres = ipasn.get(user_ip) or {}
    asn = f"{asnres.get('autonomous_system_number','unknown')}" if asnres else "unknown"
    asn_dscr = f"{asnres.get('autonomous_system_organization','unknown')}" if asnres else "unknown"
    asn_short = get_asn_short_name(user_ip)
    country = get_country(user_ip)
    default = get_host_base_on_asn(default_ips, asn_short).replace(".ircf.space", "")
    return f'{user_ip} {country} {asn} {asn_short} {"ERROR" if asn_short=="unknown" else ""} fullname={asn_dscr} default:{default}'


def get_real_user_ip():
    user_ip = request.remote_addr
    for header in ['CF-Connecting-IP', 'ar-real-ip', 'X-Forwarded-For', "X-Real-IP"]:
        if header in request.headers:
            user_ip = request.headers.get(header)
            break

    return user_ip


def get_host_base_on_asn(ips, asn_short):
    if type(ips) == str:
        ips = re.split('[ \t\r\n;,]+', ips.strip())
    valid_hosts = [ip for ip in ips if len(ip) > 5]

    if len(ips) % 2 != 0 or len(valid_hosts) == 0:
        flash(_("Error! auto cdn ip can not be find, please contact admin."))
        if len(valid_hosts) == 0:
            return

    all_hosts = []
    for i in range(0, len(ips), 2):
        if asn_short == ips[i+1]:
            # print("selected ",ips[i],ips[i+1])
            all_hosts.append(ips[i])

    selected = random.sample(valid_hosts, 1)[0]
    if len(all_hosts):
        selected = random.sample(all_hosts, 1)[0]

    return selected


def get_clean_ip(ips, resolve=False, default_asn=None):
    if not ips:
        ips = default_ips

    ips = re.split('[ \t\r\n;,]+', ips.strip())
    user_ip = get_real_user_ip()
    asn_short = get_asn_short_name(user_ip)
    country = get_country(user_ip)
    # print("Real user ip",get_real_user_ip_debug(), user_ip,asn_short,country)
    is_morteza_format = any([format for format in asn_map.values() if format in ips])
    # print("IPs",ips)
    if is_morteza_format:
        if country.lower() != hconfig(ConfigEnum.country) and default_asn:
            asn_short = default_asn
        selected_server = get_host_base_on_asn(ips, asn_short)
    else:
        selected_server = random.sample(ips, 1)[0]
    # print("selected_server",selected_server)
    if resolve:
        from hiddifypanel.panel import hiddify
        selected_server = hiddify.get_domain_ip(selected_server) or selected_server
    return selected_server



from . import bot

from telebot import types
from hiddifypanel.models import *
import datetime
from hiddifypanel.panel.database import db


@bot.message_handler(commands=['start'], func=lambda message: "admin" in message.text)
def send_welcome(message):
    text = message.text
    # print("dddd",text)
    uuid = text.split()[1].split("_")[1] if len(text.split()) > 1 else None
    if uuid:
        admin_user = get_admin_user_db(uuid)
        if admin_user:
            admin_user.telegram_id = message.chat.id
            db.session.commit()
            start_admin(message)
            return
    bot.reply_to(message, "error")


def start_admin(message):

    bot.reply_to(message, "Welcome to admin bot. Choose your action", reply_markup=admin_keyboard_main())


def get_admin_by_tgid(message):

    tgid = message.chat.id
    return AdminUser.query.filter(AdminUser.telegram_id == tgid).first()


def admin_keyboard_main():

    return types.InlineKeyboardMarkup(keyboard=[[
        types.InlineKeyboardButton(
            text="Create Package",
            callback_data=f'create_package'
        )
    ]
    ]
    )


def admin_keyboard_gig(old_action):
    def keyboard(gig):
        return types.InlineKeyboardButton(
            text=f"{gig} GB",
            callback_data=f"{old_action} {gig}"
        )
    return types.InlineKeyboardMarkup(keyboard=[
        [keyboard(i) for i in range(1, 5)],
        [keyboard(5*i) for i in range(1, 5)],
        [keyboard(50*i) for i in range(1, 5)]
    ]
    )


def admin_keyboard_days(old_action):
    def keyboard(days):
        return types.InlineKeyboardButton(
            text=f"{days}",
            callback_data=f"{old_action} {days}"
        )
    return types.InlineKeyboardMarkup(keyboard=[
        [keyboard(i) for i in range(1, 16, 3)],
        [keyboard(30*i) for i in range(1, 5)]
    ]
    )


def admin_keyboard_count(old_action):
    def keyboard(count):
        return types.InlineKeyboardButton(
            text=f"{count}",
            callback_data=f"{old_action} {count}"
        )
    return types.InlineKeyboardMarkup(keyboard=[
        [keyboard(i) for i in range(1, 5)],
        [keyboard(i*5) for i in range(1, 5)],
        [keyboard(i*50) for i in range(1, 5)]
    ]
    )


def admin_keyboard_domain(old_action):
    def keyboard(domain):
        return types.InlineKeyboardButton(
            text=f"{domain.alias or domain.domain}",
            callback_data=f"{old_action} {domain.id}"
        )
    return types.InlineKeyboardMarkup(keyboard=[
        [keyboard(d)]
        for d in get_panel_domains()
    ]
    )


@bot.callback_query_handler(func=lambda call: call.data.startswith(f'create_package'))
def create_package(call):  # <- passes a CallbackQuery type object to your function
    admin = get_admin_by_tgid(call.message)
    if not (admin):
        return
    from . import Usage
    try:
        splt = call.data.split(" ")
        if len(splt) == 1:
            new_text = "package size? حجم بسته؟"
            bot.edit_message_text(new_text, call.message.chat.id, call.message.message_id, reply_markup=admin_keyboard_gig(call.data))
            bot.answer_callback_query(call.id, text="Ok", show_alert=False, cache_time=1)
        elif len(splt) == 2:
            new_text = "package days? چند روز؟"
            bot.edit_message_text(new_text, call.message.chat.id, call.message.message_id, reply_markup=admin_keyboard_days(call.data))
            bot.answer_callback_query(call.id, text="Ok", show_alert=False, cache_time=1)
        elif len(splt) == 3:
            new_text = "How many? چه تعداد؟"
            bot.edit_message_text(new_text, call.message.chat.id, call.message.message_id, reply_markup=admin_keyboard_count(call.data))
            bot.answer_callback_query(call.id, text="Ok", show_alert=False, cache_time=1)
        elif len(splt) == 4:
            new_text = "Domain? دامنه؟"
            bot.edit_message_text(new_text, call.message.chat.id, call.message.message_id, reply_markup=admin_keyboard_domain(call.data))
            bot.answer_callback_query(call.id, text="Ok", show_alert=False, cache_time=1)
        elif len(splt) == 5:
            gig = int(splt[1])
            days = int(splt[2])
            count = int(splt[3])
            domain = int(splt[4])
            new_text = "Wait... لطفا صبر کنید"
            bot.edit_message_text(new_text, call.message.chat.id, call.message.message_id, reply_markup=None)
            DT = (ParentDomain if hconfig(ConfigEnum.is_parent) else Domain)
            domain = DT.query.filter(DT.id == domain).first()
            for i in range(1, count+1):
                user = User(package_days=days, usage_limit_GB=gig, name=f"auto {i} {datetime.date.today()}", added_by=admin.id)
                db.session.add(user)
                db.session.commit()
                # bot.send_message(call.message.chat.id,f"Days: {days}     Limit: {gig}GB     #{i}\n\n https://{domain.domain}/{hconfig(ConfigEnum.proxy_path)}/{user.uuid}/",reply_markup=Usage.user_keyboard(user.uuid))
                bot.send_message(call.message.chat.id, Usage.get_usage_msg(user.uuid, domain), reply_markup=Usage.user_keyboard(user.uuid))

            # db.session.commit()
            new_text = "Finished..."
            bot.edit_message_text(new_text, call.message.chat.id, call.message.message_id, reply_markup=admin_keyboard_main())
            bot.answer_callback_query(call.id, text="Ok", show_alert=False, cache_time=1)
            from hiddifypanel.panel import usage
            usage.update_local_usage()
            hiddify.quick_apply_users()

    except Exception as e:
        print(e)
        # import traceback
        # traceback.print_stack()
        # new_text=f"Error {e}"
        # bot.edit_message_text(new_text, call.message.chat.id, call.message.message_id,reply_markup=admin_keyboard_main())
        # bot.answer_callback_query(call.id,cache_time =1)



import datetime
import uuid as uuid_mod
from sqlalchemy_serializer import SerializerMixin
from dateutil import relativedelta
from hiddifypanel.panel.database import db
from enum import auto
from strenum import StrEnum
from flask_babelex import gettext as __
from flask_babelex import lazy_gettext as _


class UserMode(StrEnum):
    """
    The "UserMode" class is an enumeration that defines five possible modes: "no_reset", "monthly", "weekly",
    "daily", and "disable". These modes represent different settings that can be applied to a user account,
    such as the frequency at which data is reset or whether the account is currently disabled. The class is
    implemented using the "StrEnum" base class and the "auto()" function to generate unique values for each mode.
    """
    no_reset = auto()
    monthly = auto()
    weekly = auto()
    daily = auto()

    # disable = auto()


class UserDetail(db.Model, SerializerMixin):
    id = db.Column(db.Integer, primary_key=True, autoincrement=True)
    user_id = db.Column(db.Integer, db.ForeignKey('user.id'), default=0, nullable=False)
    child_id = db.Column(db.Integer, db.ForeignKey('child.id'), default=0, nullable=False)
    last_online = db.Column(db.DateTime, nullable=False, default=datetime.datetime.min)
    current_usage_GB = db.Column(db.Numeric(6, 9, asdecimal=False), default=0, nullable=False)
    connected_ips = db.Column(db.String(512), default='', nullable=False)

    @property
    def ips(self):
        return [] if not self.connected_ips else self.connected_ips.split(",")


class User(db.Model, SerializerMixin):
    """
    This is a model class for a user in a database that includes columns for their ID, UUID, name, online status,
    account expiration date, usage limit, package days, mode, start date, current usage, last reset time, and comment.
    """
    id = db.Column(db.Integer, primary_key=True, autoincrement=True)
    uuid = db.Column(db.String(36), default=lambda: str(uuid_mod.uuid4()), nullable=False, unique=True)
    name = db.Column(db.String(512), nullable=False, default='')
    last_online = db.Column(db.DateTime, nullable=False, default=datetime.datetime.min)
    # removed
    expiry_time = db.Column(db.Date, default=datetime.date.today() + relativedelta.relativedelta(months=6))
    usage_limit_GB = db.Column(db.Numeric(6, 9, asdecimal=False), default=1000, nullable=False)
    package_days = db.Column(db.Integer, default=90, nullable=False)
    mode = db.Column(db.Enum(UserMode), default=UserMode.no_reset, nullable=False)
    monthly = db.Column(db.Boolean, default=False)  # removed
    start_date = db.Column(db.Date, nullable=True)
    current_usage_GB = db.Column(db.Numeric(6, 9, asdecimal=False), default=0, nullable=False)
    last_reset_time = db.Column(db.Date, default=datetime.date.today())
    comment = db.Column(db.String(512))
    telegram_id = db.Column(db.String(512))
    added_by = db.Column(db.Integer, db.ForeignKey('admin_user.id'), default=0)
    max_ips = db.Column(db.Integer, default=1000, nullable=False)
    details = db.relationship('UserDetail', cascade="all,delete", backref='user',    lazy='dynamic',)
    enable = db.Column(db.Boolean, default=True, nullable=False)
    ed25519_private_key = db.Column(db.String(500))
    ed25519_public_key = db.Column(db.String(100))

    @property
    def remaining_days(self):
        return remaining_days(self)

    @property
    def is_active(self):
        return is_user_active(self)

    @property
    def ips(self):
        res = {}
        for detail in UserDetail.query.filter(UserDetail.user_id == self.id):
            for ip in detail.ips:
                res[ip] = 1
        return list(res.keys())

    @staticmethod
    def by_id(user_id):
        """
        Retrieves a user from the database by their ID.
        """
        return User.query.get(user_id)

    @staticmethod
    def by_uuid(user_uuid, create=False):
        """
        Retrieves a user from the database by their UUID.
        """
        dbuser = User.query.filter_by(uuid=user_uuid).first()
        if not dbuser:
            dbuser = User(uuid=user_uuid)
            db.session.add(dbuser)
        return dbuser

    @staticmethod
    def from_dict(data):
        """
        Returns a new User object created from a dictionary.
        """

        return User(
            name=data.get('name', ''),
            expiry_time=data.get('expiry_time', datetime.date.today() + relativedelta.relativedelta(months=6)),
            usage_limit_GB=data.get('usage_limit_GB', 1000),
            package_days=data.get('package_days', 90),
            mode=UserMode[data.get('mode', 'no_reset')],
            monthly=data.get('monthly', False),
            start_date=data.get('start_date', None),
            current_usage_GB=data.get('current_usage_GB', 0),
            last_reset_time=data.get('last_reset_time', datetime.date.today()),
            comment=data.get('comment', None),
            telegram_id=data.get('telegram_id', None),
            added_by=data.get('added_by', 0)
        )

    # @staticmethod
    # def from_dict(data):

    #     allowed_fields = {'name', 'expiry_time', 'usage_limit_GB', 'package_days', 'mode', 'monthly', 'start_date',
    #                       'current_usage_GB', 'comment', 'telegram_id', 'added_by_uuid'}
    #     filtered_data = {k: v for k, v in data.items() if k in allowed_fields}
    #     if 'mode' in filtered_data:
    #         filtered_data['mode'] = UserMode[filtered_data['mode']]
    #     if 'expiry_time' in filtered_data:
    #         filtered_data['expiry_time'] = datetime.fromisoformat(filtered_data['expiry_time']).date()
    #     if 'start_date' in filtered_data and filtered_data['start_date']:
    #         filtered_data['start_date'] = datetime.fromisoformat(filtered_data['start_date']).date()
    #     return User(**filtered_data)
    def to_dict(d):
        from hiddifypanel.panel import hiddify
        return {
            'uuid': d.uuid,
            'name': d.name,
            'last_online': hiddify.time_to_json(d.last_online),
            'usage_limit_GB': d.usage_limit_GB,
            'package_days': d.package_days,
            'mode': d.mode,
            'start_date': hiddify.date_to_json(d.start_date),
            'current_usage_GB': d.current_usage_GB,
            'last_reset_time': hiddify.date_to_json(d.last_reset_time),
            'comment': d.comment,
            'added_by_uuid': d.admin.uuid,
            'telegram_id': d.telegram_id,
            'ed25519_private_key': d.ed25519_private_key,
            'ed25519_public_key': d.ed25519_public_key
        }


def is_user_active(u):
    """
    The "is_user_active" function checks if the input user object "u" is active by verifying if their mode is not
    "disable", if their usage limit hasn't been exceeded, and if there are remaining days on their account. The
    function returns a boolean value indicating whether the user is active or not.
    """
    is_active = True
    if not u:
        is_active = False
    elif not u.enable:
        is_active = False
    elif u.usage_limit_GB < u.current_usage_GB:
        is_active = False
    elif remaining_days(u) < 0:
        is_active = False
    elif len(u.ips) > max(3, u.max_ips):
        is_active = False
    return is_active


def remaining_days(u):
    """
    The "remaining_days" function calculates the number of days remaining for a user's account package based on the
    current date and the user's start date. The function returns the remaining days as an integer value. If the start
    date is not available, the function returns the total package days.
    """
    res = -1
    if u.package_days is None:
        res = -1
    elif u.start_date:
        # print(datetime.date.today(), u.start_date,u.package_days, u.package_days - (datetime.date.today() - u.start_date).days)
        res = u.package_days - (datetime.date.today() - u.start_date).days
    else:
        # print("else",u.package_days )
        res = u.package_days
    return min(res, 10000)


package_mode_dic = {
    UserMode.daily: 1,
    UserMode.weekly: 7,
    UserMode.monthly: 30
}


def days_to_reset(user):
    """
    The "days_to_reset" function calculates the number of days until the user's data usage is reset, based on the
    user's start date and mode of usage. The function returns the remaining days as an integer value. If the start
    date is not available, the function returns the total days for the user's mode.
    """
    # print("start_date",user.start_date, "pack",package_mode_dic.get(user.mode,10000), "total",(datetime.date.today()-user.start_date).days)
    # if user.mode==UserMode.daily:
    #     return 0
    if user.start_date:
        days = package_mode_dic.get(user.mode, 10000)-(datetime.date.today()-user.start_date).days % package_mode_dic.get(user.mode, 10000)
    else:
        days = package_mode_dic.get(user.mode, 10000)
    return max(-100000, min(days, 100000))


def user_should_reset(user):
    """
    The "days_to_reset" function calculates the number of days until the user's data usage is reset, based on the
    user's start date and mode of usage. The function returns the remaining days as an integer value. If the start
    date is not available, the function returns the total days for the user's mode.
    """
    # print("start_date",user.start_date, "pack",package_mode_dic.get(user.mode,10000), "total",(datetime.date.today()-user.start_date).days)
    # if user.mode==UserMode.daily:
    #     return 0
    res = True
    if not user.last_reset_time:
        res = True
    elif not user.start_date or (datetime.date.today()-user.last_reset_time).days == 0:
        res = False
    else:
        res = ((datetime.date.today()-user.start_date).days % package_mode_dic.get(user.mode, 10000)) == 0

    return res


def user_by_uuid(uuid):
    return User.query.filter(User.uuid == uuid).first()


def user_by_id(id):
    return User.query.filter(User.id == id).first()


def add_or_update_user(commit=True, **user):
    # if not is_valid():return
    from hiddifypanel.panel import hiddify
    dbuser = User.by_uuid(user['uuid'], create=True)

    if user.get('added_by_uuid'):
        from .admin import get_admin_by_uuid
        admin = get_admin_by_uuid(user.get('added_by_uuid'), create=True)
        dbuser.added_by = admin.id
    else:
        dbuser.added_by = 1

    if user.get('expiry_time', ''):
        last_reset_time = hiddify.json_to_date(user.get('last_reset_time', '')) or datetime.date.today()

        expiry_time = hiddify.json_to_date(user['expiry_time'])
        dbuser.start_date = last_reset_time
        dbuser.package_days = (expiry_time-last_reset_time).days

    elif 'package_days' in user:
        dbuser.package_days = user['package_days']
        if user.get('start_date', ''):
            dbuser.start_date = hiddify.json_to_date(user['start_date'])
        else:
            dbuser.start_date = None
    dbuser.current_usage_GB = user['current_usage_GB']

    dbuser.usage_limit_GB = user['usage_limit_GB']
    dbuser.name = user.get('name') or ''
    dbuser.comment = user.get('comment', '')
    dbuser.enable = user.get('enable', True)
    if user.get('ed25519_private_key', ''):
        dbuser.ed25519_private_key = user.get('ed25519_private_key', '')
        dbuser.ed25519_public_key = user.get('ed25519_public_key', '')
    if not dbuser.ed25519_private_key:
        priv, publ = hiddify.get_ed25519_private_public_pair()
        dbuser.ed25519_private_key = priv
        dbuser.ed25519_public_key = publ

    mode = user.get('mode', UserMode.no_reset)
    if mode == 'disable':
        mode = UserMode.no_reset
        dbuser.enable = False

    dbuser.mode = mode

    dbuser.telegram_id = user.get('telegram_id')

    dbuser.last_online = hiddify.json_to_time(user.get('last_online')) or datetime.datetime.min

    if commit:
        db.session.commit()


def bulk_register_users(users=[], commit=True, remove=False):
    for u in users:
        add_or_update_user(commit=False, **u)
    if remove:
        dd = {u.uuid: 1 for u in users}
        for d in User.query.all():
            if d.uuid not in dd:
                db.session.delete(d)
    if commit:
        db.session.commit()



import click


from hiddifypanel.panel.database import db
from hiddifypanel.panel.init_db import init_db
from hiddifypanel.models import *

from hiddifypanel.panel import hiddify, usage
import random
import uuid
import urllib
import string
from dateutil import relativedelta
import datetime


def drop_db():
    """Cleans database"""
    db.drop_all()


def downgrade():
    if (hconfig(ConfigEnum.db_version) >= "49"):
        set_hconfig(ConfigEnum.db_version, 42, commit=False)
        StrConfig.query.filter(StrConfig.key.in_([ConfigEnum.tuic_enable, ConfigEnum.tuic_port, ConfigEnum.hysteria_enable,
                               ConfigEnum.hysteria_port, ConfigEnum.ssh_server_enable, ConfigEnum.ssh_server_port, ConfigEnum.ssh_server_redis_url])).delete()
        Proxy.query.filter(Proxy.l3.in_([ProxyL3.ssh, ProxyL3.h3_quic, ProxyL3.custom])).delete()
        db.session.commit()


def backup():
    dbdict = hiddify.dump_db_to_dict()
    import json
    import os
    os.makedirs('backup', exist_ok=True)
    dst = f'backup/{datetime.datetime.now().strftime("%Y_%m_%d__%H_%M_%S")}.json'
    with open(dst, 'w') as fp:
        json.dump(dbdict, fp, indent=4, sort_keys=True, default=str)

    if hconfig(ConfigEnum.telegram_bot_token):
        for admin in AdminUser.query.filter(AdminUser.mode == AdminMode.super_admin, AdminUser.telegram_id != None).all():
            from hiddifypanel.panel.commercial.telegrambot import bot
            with open(dst, 'rb') as document:
                caption = ("Backup \n"+admin_links())
                bot.send_document(admin.telegram_id, document, visible_file_name=dst.replace("backup/", ""), caption=caption[:min(len(caption), 1000)])


def all_configs():
    import json
    valid_users = [u.to_dict() for u in User.query.filter((User.usage_limit_GB > User.current_usage_GB)).all() if is_user_active(u)]

    configs = {
        "users": valid_users,
        "domains": [u.to_dict() for u in Domain.query.all() if "*" not in u.domain],
        # "parent_domains": [hiddify.parent_domain_dict(u) for u in ParentDomain.query.all()],
        "hconfigs": get_hconfigs()
    }
    for d in configs['domains']:
        d['domain'] = d['domain'].lower()
        # del d['domain']['show_domains']

    def_user = None if len(User.query.all()) > 1 else User.query.filter(User.name == 'default').first()
    domains = Domain.query.all()
    sslip_domains = [d.domain for d in domains if "sslip.io" in d.domain]

    configs['hconfigs']['first_setup'] = def_user != None and len(sslip_domains) > 0
    # configs

    print(json.dumps(configs))


def update_usage():
    print(usage.update_local_usage())


def test():
    print(ConfigEnum("auto_update1"))


def admin_links():
    proxy_path = hconfig(ConfigEnum.proxy_path)

    admin_secret = get_super_admin_secret()
    server_ip = hiddify.get_ip(4)
    admin_links = f"Not Secure (do not use it- only if others not work):\n   http://{server_ip}/{proxy_path}/{admin_secret}/admin/\n"

    domains = get_panel_domains()
    admin_links += f"Secure:\n"
    if not any([d for d in domains if 'sslip.io' not in d.domain]):
        admin_links += f"   (not signed) https://{server_ip}/{proxy_path}/{admin_secret}/admin/\n"

    # domains=[*domains,f'{server_ip}.sslip.io']
    for d in domains:
        admin_links += f"   https://{d.domain}/{proxy_path}/{admin_secret}/admin/\n"

    print(admin_links)
    return admin_links


def admin_path():
    proxy_path = hconfig(ConfigEnum.proxy_path)
    admin = Admin.query.filter(Admin.mode == AdminMode.super_admin).first()
    if not admin:
        db.session.add(Admin(mode=AdminMode.super_admin))
        db.session.commit()
        admin = Admin.query.filter(Admin.mode == AdminMode.super_admin).first()

    admin_secret = admin.uuid
    print(f"/{proxy_path}/{admin_secret}/admin/")


def init_app(app):
    # add multiple commands in a bulk
    # print(app.config['SQLALCHEMY_DATABASE_URI'] )
    for command in [init_db, drop_db, all_configs, update_usage, test, admin_links, admin_path, backup, downgrade]:
        app.cli.add_command(app.cli.command()(command))

    @app.cli.command()
    @click.option("--domain", "-d")
    def add_domain(domain):
        table = ParentDomain if hconfig(ConfigEnum.is_parent) else Domain

        if table.query.filter(table.domain == domain).first():
            return "Domain already exist."
        d = table(domain=domain)
        if not hconfig(ConfigEnum.is_parent):
            d.mode = DomainType.direct
        db.session.add(d)
        db.session.commit()
        return "success"

    @app.cli.command()
    @click.option("--admin_secret", "-a")
    def set_admin_secret(admin_secret):
        StrConfig.query.filter(StrConfig.key == ConfigEnum.admin_secret).update({'value': admin_secret})
        db.session.commit()
        return "success"

    @app.cli.command()
    @click.option("--key", "-k")
    @click.option("--val", "-v")
    def set_setting(key, val):
        old_hconfigs = get_hconfigs()
        hiddify.add_or_update_config(key=key, value=val)

        return "success"

    @app.cli.command()
    @click.option("--config", "-c")
    def import_config(config):
        next10year = datetime.date.today() + relativedelta.relativedelta(years=10)
        data = []
        if "USER_SECRET" in config:
            secrets = config["USER_SECRET"].split(";")
            for i, s in enumerate(secrets):
                data.append(User(name=f"default {i}", uuid=uuid.UUID(s), usage_limit_GB=9000, package_days=3650))

        if "MAIN_DOMAIN" in config:
            domains = config["MAIN_DOMAIN"].split(";")
            for i, d in enumerate(domains):
                if not Domain.query.filter(Domain.domain == d).first():
                    data.append(Domain(domain=d, mode=DomainType.direct),)

        strmap = {
            "TELEGRAM_FAKE_TLS_DOMAIN": ConfigEnum.telegram_fakedomain,
            "TELEGRAM_SECRET": ConfigEnum.shared_secret,
            "SS_FAKE_TLS_DOMAIN": ConfigEnum.ssfaketls_fakedomain,
            "FAKE_CDN_DOMAIN": ConfigEnum.domain_fronting_domain,
            "BASE_PROXY_PATH": ConfigEnum.proxy_path,
            "ADMIN_SECRET": ConfigEnum.admin_secret,
            "TELEGRAM_AD_TAG": ConfigEnum.telegram_adtag
        }
        boolmap = {
            "ENABLE_SS": ConfigEnum.ssfaketls_enable,
            "ENABLE_TELEGRAM": ConfigEnum.telegram_enable,
            "ENABLE_VMESS": ConfigEnum.vmess_enable,
            # "ENABLE_MONITORING":ConfigEnum.ssfaketls_enable,
            "ENABLE_FIREWALL": ConfigEnum.firewall,
            "ENABLE_NETDATA": ConfigEnum.netdata,
            "ENABLE_HTTP_PROXY": ConfigEnum.http_proxy_enable,
            "ALLOW_ALL_SNI_TO_USE_PROXY": ConfigEnum.allow_invalid_sni,
            "ENABLE_AUTO_UPDATE": ConfigEnum.auto_update,
            "ENABLE_SPEED_TEST": ConfigEnum.speed_test,
            "BLOCK_IR_SITES": ConfigEnum.block_iran_sites,
            "ONLY_IPV4": ConfigEnum.only_ipv4
        }

        for k in config:
            if k in strmap:
                if hconfig(strmap[k]) is None:
                    data.append(StrConfig(key=strmap[k], value=config[k]))
                else:
                    StrConfig.query.filter(StrConfig.key == strmap[k]).update({
                        'value': config[k]
                    })
            if k in boolmap:
                if hconfig(boolmap[k]) is None:
                    data.append(BoolConfig(key=boolmap[k], value=config[k]))
                else:
                    BoolConfig.query.filter(BoolConfig.key == strmap[k]).update({
                        'value': config[k]
                    })
        if len(data):
            db.session.bulk_save_objects(data)
        db.session.commit()
